from pathlib import Path
import numpy as np
from scipy import signal as sc_signal
import pandas as pd

from pyadlml.constants import START_TIME, END_TIME, TIME, VALUE, DEVICE, BOOL, NUM, CAT
from pyadlml.dataset.util import df_difference, infer_dtypes

CORRECTION_TS = 'correction_ts'
CORRECTION_ONOFF_INCONS = 'correction_on_off_inconsistency'
COLS_DEV = [TIME, DEVICE, VALUE]

"""
    df_devices:
        Standard representation for device dataframes.
        Exc: 
           | time      | device    | state
        ------------------------------------
         0  | timestamp | dev_name  |   1

    Representation 2:
        is used to calculate statistics for binary devices more easily. 
        Some computation is easier using this format
           | start_time | end_time   | device    | state
        --------------------------------------------------
         0 | timestamp   | timestamp | dev_name  | state_0
 
 """


def _create_devices(dev_list: list, index=None):
    """
    creates an empty device dataframe
    """
    if index is not None:
        return pd.DataFrame(columns=dev_list, index=index)
    else:
        return pd.DataFrame(columns=dev_list)


def _check_devices_sequ_order(df: pd.DataFrame):
    """
    iterate pairwise through each select device an check if the 
    sequential order in time is inconsistent

    Parameters
    ----------
    df : pd.DataFrame
        device representation 1  with columns [start_time, end_time, devices]
    """
    dev_list = df[DEVICE].unique()
    no_errors = True
    for dev in dev_list:
        df_d = df[df[DEVICE] == dev]
        for i in range(1, len(df_d)):
            st_j = df_d.iloc[i - 1].start_time
            et_j = df_d.iloc[i - 1].end_time
            st_i = df_d.iloc[i].start_time
            # et_i = df_d.iloc[i].end_time
            # if the sequential order is violated return false
            if not (st_j < et_j) or not (et_j < st_i):
                print('~' * 50)
                if st_j >= et_j:
                    # raise ValueError('{}; st: {} >= et: {} |\n {} '.format(i-1, st_j, et_j, df_d.iloc[i-1]))
                    print('{}; st: {} >= et: {} |\n {} '.format(i - 1, st_j, et_j, df_d.iloc[i - 1]))
                if et_j >= st_i:
                    # raise ValueError('{},{}; et: {} >= st: {} |\n{}\n\n{}'.format(i-1,i, et_j, st_i, df_d.iloc[i-1], df_d.iloc[i]))
                    print('{},{}; et: {} >= st: {} |\n{}\n\n{}'.format(i - 1, i, et_j, st_i, df_d.iloc[i - 1],
                                                                       df_d.iloc[i]))
                no_errors = False
    return no_errors


def _is_dev_rep2(df: pd.DataFrame):
    """
    """
    if not START_TIME in df.columns or not END_TIME in df.columns \
            or not DEVICE in df.columns or len(df.columns) != 3:
        return False
    return True


def is_device_df(df: pd.DataFrame) -> bool:
    """ Checks if a dataframe is a valid device dataframe.
    """
    return DEVICE in df.columns \
           and VALUE in df.columns \
           and TIME in df.columns \
           and len(df.columns) == 3

def read_device_df(path_to_file: Path) -> None:
    devs = pd.read_csv(path_to_file)
    devs[TIME] = pd.to_datetime(devs[TIME])
    return devs

def write_device_df(path_to_file: Path, df: pd.DataFrame) -> None:
    df.to_csv(path_to_file, index=False)

def get_index_matching_rows(df_devs, rows, tolerance='1ms'):

    if isinstance(rows, list):
        df = pd.DataFrame(rows, columns=COLS_DEV)
        df[TIME] = pd.to_datetime(df[TIME], errors='coerce', dayfirst=True)
    else:
        print('went here')
        df = rows

    assert isinstance(df, pd.DataFrame)

    tol = pd.Timedelta(tolerance)
    idxs = []
    for _, row in df.iterrows():
        mask_st = (row[TIME]-tol < df_devs[TIME])\
                & (df_devs[TIME] < row[TIME]+tol)
        mask_dev = (df_devs[DEVICE] == row[DEVICE])
        res = df_devs[mask_st & mask_dev].index.values
        assert len(res) <= 1
        if len(res) == 1:
            idxs.append(*res)
        if len(res) == 0:
            print('Warning!!! Tried to delete device but could not')
    return idxs






def device_events_to_states(df_devs: pd.DataFrame, start_time=None, end_time=None, extrapolate_states=False):
    """ Transforms device dataframe from an event representation into a state representation,
        
    Parameters
    ----------
    df_devs : pd.DataFrame
        In event representation, a dataframe with columns (time, device, value)
        example row: [2008-02-25 00:20:14, Freezer, False]
    extrapolate_states : Boolean, default=False
        Whether boolean devices should add extra states for the first occurring events
    start_time : pd.Timestamp, str or default=None
        The start time from which to
    end_time : pd.Timestamp, str or default=None
        The start time from which to

    Returns
    -------
    df : pd.DataFrame
        In state representation with columns (start time, end_time, device, value)
        example row: [2008-02-25 00:20:14, 2008-02-25 00:22:14, Freezer, True]

    df, lst
    """
    epsilon = '1ns'

    df = df_devs.copy() \
        .reset_index(drop=True) \
        .sort_values(TIME)



    dtypes = infer_dtypes(df)

    first_timestamp = df[TIME].iloc[0] if start_time is None else pd.Timestamp(start_time)
    last_timestamp  = df[TIME].iloc[-1] if end_time is None else pd.Timestamp(end_time)

    if extrapolate_states:
        # create additional rows in order to compensate for the state duration of the first event
        # to the selected event for boolean devices
        eps = pd.Timedelta(epsilon)

        # Prevents boolean cast warning when extrapolatiing states
        df[VALUE] = df[VALUE].astype(object)

        for dev in dtypes[BOOL]:
            df_dev = df[df[DEVICE] == dev]
            first_row = df_dev.iloc[0].copy()
            if first_row[TIME] != first_timestamp and first_row[TIME] - first_timestamp > pd.Timedelta('1s'):
                first_row[VALUE] = not first_row[VALUE]
                first_row[TIME] = first_timestamp + eps
                df = pd.concat([df, first_row.to_frame().T], axis=0, ignore_index=True)
            eps += pd.Timedelta(epsilon)

    df = df.sort_values(TIME).reset_index(drop=True)

    lst_cat_or_bool = dtypes[CAT] + dtypes[BOOL]
    res = pd.DataFrame(columns=[START_TIME, END_TIME, DEVICE, VALUE])

    if lst_cat_or_bool:
        df_cat_bool = df[df[DEVICE].isin(lst_cat_or_bool)].copy()
        df_cat_bool = df_cat_bool.rename(columns={TIME: START_TIME})
        df_cat_bool[END_TIME] = pd.NaT
        for dev in lst_cat_or_bool:
            mask = (df_cat_bool[DEVICE] == dev)
            df_cat_bool.loc[mask, END_TIME] = df_cat_bool.loc[mask, START_TIME].shift(-1)
        if extrapolate_states:
            df_cat_bool = df_cat_bool.fillna(last_timestamp)
        else:
            df_cat_bool = df_cat_bool.dropna()
        res = pd.concat([res, df_cat_bool])

    if dtypes[NUM]:
        df_num = df[df[DEVICE].isin(dtypes[NUM])].copy()
        df_num = df_num.rename(columns={TIME: START_TIME})
        df_num[END_TIME] = df_num[START_TIME].copy()
        res = res.append(df_num)

    res[START_TIME] = pd.to_datetime(res[START_TIME])
    res[END_TIME] = pd.to_datetime(res[END_TIME])

    return res


def device_boolean_on_states_to_events(df_devs_states: pd.DataFrame):
    """
    Parameters
    ----------
    df_devs_states : pd.DataFrame
        A table with the columns (start time, end_time, device)
        All states are assumed to be 'on'/True

    Returns
    -------
    df_devs : (pd.DataFrame)
        rep1: columns are (time, device, val)
        example row: [2008-02-25 00:20:14, Freezer, False]
    """
    # copy devices to new dfs
    # one with all values but start time and other way around
    df_start = df_devs_states.copy().drop(columns=END_TIME)
    df_end = df_devs_states.copy().drop(columns=START_TIME)

    # Set values at the end time to zero because this is the time a device turns off
    df_start[VALUE] = True
    df_end[VALUE] = False

    # Rename column 'End Time' and 'Start Time' to 'Time'
    df_start.rename(columns={START_TIME: TIME}, inplace=True)
    df_end.rename(columns={END_TIME: TIME}, inplace=True)

    df = pd.concat([df_end, df_start]).sort_values(TIME) \
        .reset_index(drop=True)
    return df


def device_states_to_events(df_devs_states: pd.DataFrame):
    """
    Parameters
    ----------
    df_devs_states : pd.DataFrame
        A table with the columns (start time, end_time, device, val).

    Returns
    -------
    df_devs : (pd.DataFrame)
        The table columns are (time, device, val)

    Example
    -------
    >>> device_states_to_events(df_devs_states)
        time,               device,   value
        2008-02-25 00:20:14, Freezer, False
    """

    # copy devices to new dfs
    # one with all values but start time and other way around
    # tmp = df_devs_states.copy().rename(columns={START_TIME: TIME})
    # res = []
    # if dtypes[BOOL]:
    #    mask_bool_true = df_devs_states[DEVICE].isin(dtypes[BOOL]) \
    #                    & (df_devs_states[VAL] == True)
    #    df = df_devs_states.loc[mask_bool_true, [START_TIME, END_TIME, VAL]]
    #    df_devs_boolean = device_boolean_on_states_to_events(df.copy())
    #    res.append(df_devs_boolean)
    df = df_devs_states.copy()
    df.rename(columns={START_TIME: TIME}, inplace=True)
    df.drop(columns=END_TIME, inplace=True)
    return df


def correct_device_ts_duplicates(df: pd.DataFrame):
    """
    remove devices that went on and off at the same time. And add a microsecond
    to devices that trigger on the same time
    Parameters
    ----------
    df : pd.DataFrame
        Devices in representation 1; columns [time, device, value]
    """
    eps = pd.Timedelta('10ms')

    try:
        df[TIME]
    except KeyError:
        df = df.reset_index()

    # remove device if it went on and off at the same time    
    dup_mask = df.duplicated(subset=[TIME, DEVICE], keep=False)
    df = df[~dup_mask]

    df = df.reset_index(drop=True)

    dup_mask = df.duplicated(subset=[TIME], keep=False)
    duplicates = df[dup_mask]
    uniques = df[~dup_mask]

    # for every pair of duplicates add a millisecond on the second one
    duplicates = duplicates.reset_index(drop=True)
    sp = duplicates[TIME] + eps
    mask_p = (duplicates.index % 2 == 0)

    duplicates[TIME] = duplicates[TIME].where(mask_p, sp)

    # concatenate and sort the dataframe 
    uniques = uniques.set_index(TIME)
    duplicates = duplicates.set_index(TIME)
    df = pd.concat([duplicates, uniques], sort=True)

    # set the time as index again
    df = df.sort_values(TIME).reset_index(drop=False)

    return df


def _has_timestamp_duplicates(df: pd.DataFrame):
    """ check whether there are duplicates in timestamp present
    Parameters
    ----------
    df : pd.DataFrame
        data frame representation 1: [time, device, val]
    """
    df = df.copy()
    try:
        dup_mask = df.duplicated(subset=[TIME], keep=False)
    except KeyError:
        df = df.reset_index()
        dup_mask = df.duplicated(subset=[TIME], keep=False)
    return dup_mask.sum() > 0


def split_devices_binary(df: pd.DataFrame):
    """ separate binary devices and non-binary devices
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe in device representation 1
    Returns
    -------
    df_binary, df_non_binary : pd.DataFrames
        Dataframe with binary devices and dataframe without binary devices
    """
    mask_binary = df[VALUE].apply(lambda x: isinstance(x, bool))
    return df[mask_binary], df[~mask_binary]


def contains_non_binary(df: pd.DataFrame) -> bool:
    """ determines whether the the dataframes values contain non-boolean values
    These can be continuous values of categorical.
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe in device representation 1
    Returns
    -------
    boolean
    """
    return not df[VALUE].apply(lambda x: isinstance(x, bool)).all()


def correct_devices(df: pd.DataFrame, retain_correction=False) -> pd.DataFrame:
    """
    Applies correction to devices.
    1. drops all duplicates
    2. timestamps that are identical are separated by an offset
    3. for binary devices redundant reportings of the same value are dropped

    Parameters
    ----------
    df : pd.DataFrame
        either in device representation 1 or 2
    retain_correction: bool, default=False
        decides whether corrections are computed and returned or not.

    Returns
    -------
    cor_rep1 : pd.DataFrame
        Device dataframe in representation 1
    corrections : dict or None

    """
    if retain_correction:
        corrections = {}
    else:
        corrections = None

    df = df.copy()
    df = df.drop_duplicates()

    if df.empty:
        return df

    df = df.sort_values(by=TIME).reset_index(drop=True)

    df_cor = df.copy()
    # correct timestamp duplicates
    while _has_timestamp_duplicates(df_cor):
        df_cor = correct_device_ts_duplicates(df_cor)
    assert not _has_timestamp_duplicates(df_cor)

    if retain_correction:
        corrections[CORRECTION_TS] = df_difference(df, df_cor)

    df = df_cor.copy()

    if contains_non_binary(df):
        df_binary, df_non_binary = split_devices_binary(df)
        non_binary_exist = True
    else:
        df_binary = df
        non_binary_exist = False

    # correct on/off inconsistency
    if not is_on_off_consistent(df_binary):
        df_binary = correct_on_off_inconsistency(df_binary)
    assert is_on_off_consistent(df_binary)

    # join dataframes
    if non_binary_exist:
        df = pd.concat([df_binary, df_non_binary], axis=0, ignore_index=True)
    else:
        df = df_binary

    if retain_correction:
        corrections[CORRECTION_ONOFF_INCONS] = df_difference(df_cor, df)

    df = df.sort_values(by=TIME).reset_index(drop=True)

    return df, corrections


def on_off_consistent_func(df: pd.DataFrame, dev: list):
    """ compute for each device if it is on/off consistent
    Parameters
    ----------
    df : pd.DataFrame
        the whole activity dataframe
    dev : str
        a device that occurs in the dataframe
    Returns
    -------
    tupel [arg1, arg2]
        first argument is a boolean whether this device is consistent
        second is
    """
    df_dev = df[df[DEVICE] == dev].sort_values(by=TIME).reset_index(drop=True)
    first_val = df_dev[VALUE].iloc[0]
    if first_val:
        mask = np.zeros((len(df_dev)), dtype=bool)
        mask[::2] = True
    else:
        mask = np.ones((len(df_dev)), dtype=bool)
        mask[::2] = False
    return [not (df_dev[VALUE] ^ mask).sum() > 0, df_dev[[TIME, DEVICE, VALUE]]]


def is_on_off_consistent(df: pd.DataFrame):
    """ devices can only go on after they are off and vice versa. check if this is true
        for every device.
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe in representation 1.
        The dataframe must not include timestamp duplicates! When it does they can be arbitrarily reordered
        when sorting for time thus destroying the consistency.
    """

    import dask
    from dask import delayed
    lazy_results = []
    for dev in df[DEVICE].unique():
        res = dask.delayed(on_off_consistent_func)(df.copy(), dev)
        lazy_results.append(res)

    results = np.array(list(dask.compute(*lazy_results)), dtype=object)
    return results[:, 0].all()


def correct_on_off_inconsistency(df: pd.DataFrame):
    """ 
    has multiple strategies for solving the patterns:
    e.g if the preceeding value is on and the same value is occuring now delete now
    Parameters
    ----------
    df : pd.DataFrame
        device representation 3
        
    Returns
    -------
    df : pd.DataFrame
        device representation 3
    """
    from dask import delayed
    import dask

    def correct_part(df_dev):
        """ get index of rows where the previous on/off value was the same and delete row
        Parameters
        ----------
        df_dev : pd.DataFrame
            subset of the big dataframe consisting only of events for a fixed device
        """
        df_dev['same_prec'] = ~(df_dev[VALUE].shift(1) ^ df_dev[VALUE])
        df_dev.loc[0, 'same_prec'] = False  # correct shift artefact
        indices = list(df_dev[df_dev['same_prec']].index)
        df_dev = df_dev.drop(indices, axis=0)
        return df_dev[[TIME, DEVICE, VALUE]]

    df = df.copy()
    # create list of tuples e.g [(True, df_dev1), (False, df_dev2), ...]
    dev_list = df[DEVICE].unique()
    dev_df_list = []
    for devs in dev_list:
        dev_df_list.append(delayed(on_off_consistent_func)(df, devs))
    results = np.array(list(dask.compute(*dev_df_list)), dtype=object)

    # filter inconsistent devices
    incons = results[np.where(np.logical_not(results[:, 0]))[0], :][:, 1]

    corrected = []
    for part in incons:
        corrected.append(delayed(correct_part)(part))
    corr_dfs = delayed(lambda x: x)(corrected).compute()
    corr_devs = [df[DEVICE].iloc[0] for df in corr_dfs]
    tmp = [df[~df[DEVICE].isin(corr_devs)], *corr_dfs]

    return pd.concat(tmp, ignore_index=True) \
        .sort_values(by=TIME).reset_index(drop=True)


def _create_empty_dev_dataframe():
    df = pd.DataFrame(data=[], columns=[TIME, DEVICE, VALUE])
    df[TIME] = pd.to_datetime(df[TIME])
    return df


def most_prominent_categorical_values(df: pd.DataFrame):
    tmp = df
    tmp['conc'] = tmp[DEVICE] + ',' + tmp[VALUE]
    dev_list = tmp[DEVICE].unique()
    tmp = pd.concat([tmp[TIME], pd.get_dummies(tmp['conc'], dtype=int)], axis=1)
    tmp2 = tmp.copy()
    for dev in dev_list:
        dev_col_cats = []
        for col in tmp.columns:
            if dev not in col:
                continue
            dev_col_cats.append(col)

        for di in dev_col_cats:
            other = dev_col_cats[:]
            other.remove(di)
            for o in other:
                tmp2[di] = tmp2[di] - tmp[o]
    tmp2 = tmp2.set_index(TIME)
    tmp2 = tmp2.where(tmp2 != 0, np.nan)
    tmp2 = tmp2.ffill(axis=0)
    tmp2 = tmp2.where(tmp2 != -1, 0)
    tmp2 = tmp2.reset_index()
    tmp2['td'] = tmp2[TIME].shift(-1) - tmp2[TIME]

    lst = []
    for dev in dev_list:
        dev_col_cats = []
        for col in tmp.columns:
            if dev not in col:
                continue
            dev_col_cats.append(col)
        asdf = []
        max_td = pd.Timedelta('0s')
        max_cat = None
        for di in dev_col_cats:
            abc = tmp2[[di, 'td']].groupby(di).sum().reset_index()
            td_on = abc.iloc[1, 1]
            if max_td < td_on:
                max_td = td_on
                max_cat = di
            asdf.append(td_on)
        lst.append([dev, max_cat.split(',')[1]])
    ML_STATE = 'ml_state'
    df = pd.DataFrame(data=lst, columns=[DEVICE, ML_STATE])
    return df


def _get_bool_mask(df: pd.DataFrame, col: list):
    """ returns a mask where the columns are booleans"""
    if df[col].dtype == 'bool':
        # return df[col].na # TODO hack for returning
        return (df[col] == True) | (df[col] == False)
    mask = (df[col].astype(str) == 'False') | (df[col].astype(str) == 'True')
    return mask


def _get_num_mask(df, col):
    df = df.copy()
    df[col] = df[col].astype(str)
    df[col] = df[col].where(df[col] != 'True', 'stub')
    df[col] = df[col].where(df[col] != 'False', 'stub')
    num_mask = pd.to_numeric(df[col], errors='coerce').notnull()
    return num_mask


def _get_bool(df):
    def func(x):
        # check if every is one of
        if x[VALUE].dtype == 'bool':
            return True
        else:
            return ((x[VALUE].astype(str) == 'False') | (x[VALUE].astype(str) == 'True')).all()

    return df.groupby(by=[DEVICE]).filter(func)


def create_device_info_dict(df_dev: pd.DataFrame) -> dict:
    """
    Infers for each device the most likely state

    Parameters
    ----------
    df_dev : pd.DataFrame
        A device dataframe
    dtypes: pd.DataFrame
        a dataframe containing every device and its corresponding datattype

    Returns
    -------
    df : pd.DataFrame
        the result
    """
    ML_STATE = 'ml_state'
    DTYPE = 'dtype'
    df_dev = df_dev.copy()

    # extract devices of differing dtypes
    dtypes = infer_dtypes(df_dev)

    res = {}
    for key in dtypes.keys():
        for dev in dtypes[key]:
            res[dev] = {}
            res[dev][DTYPE] = key

    # get most likely binary states
    if dtypes[BOOL]:
        from pyadlml.dataset.stats.devices import state_fractions
        dsf = state_fractions(df_dev[df_dev[DEVICE].isin(dtypes[BOOL])].copy())
        for dev in dtypes[BOOL]:
            true_has_more = dsf.loc[(dsf[DEVICE] == dev) & (dsf[VALUE] == True), 'frac'].values[0] \
                            > dsf.loc[(dsf[DEVICE] == dev) & (dsf[VALUE] == False), 'frac'].values[0]
            res[dev][ML_STATE] = true_has_more

    # get most likely numerical states
    # use median for the most likely numerical state
    if dtypes[NUM]:
        df_num = df_dev.loc[df_dev[DEVICE].isin(dtypes[NUM]), [DEVICE, VALUE]].copy()
        df_num[VALUE] = pd.to_numeric(df_num[VALUE])
        res_num = df_num.groupby(by=[DEVICE]).median()
        for dev in dtypes[NUM]:
            res[dev][ML_STATE] = res_num.at[dev, VALUE]

    # get most likely categorical states
    if dtypes[CAT]:
        df_cat = df_dev[df_dev[DEVICE].isin(dtypes[CAT])].copy()
        res_cat = most_prominent_categorical_values(df_cat)
        res_cat.set_index(DEVICE, inplace=True)
        for dev in dtypes[CAT]:
            res[dev][ML_STATE] = res_cat.at[dev, ML_STATE]

    return res


def device_remove_state(df_devs, state, td, eps='0.2s'):
    """ Remove the events corresponding to a device that is a certain time
        in the specified state.

    Parameters
    ----------
    df_devs : pd.DataFrame
    state : bool
    td : str
    eps : str, default='0.2s'

    Returns
    ------
    pd.DataFrame
        The table with the specified events missing
    """
    eps = pd.Timedelta(eps)
    td = pd.Timedelta(td)

    # Note that due to extrapolate_states the very last events
    # for each device are disregarded
    df = device_events_to_states(df_devs.copy(), extrapolate_states=False)

    # Mark states that fall match the criterion
    df['diff'] = df[END_TIME] - df[START_TIME]
    df['target'] = (td - eps < df['diff']) & (df['diff'] < td + eps) \
                   & (df[VALUE] == state)

    # Remove states
    df = df[df['target'] == False]

    df = device_states_to_events(df[[START_TIME, END_TIME, DEVICE, VALUE]])
    df = correct_on_off_inconsistency(df)

    # Since the transformation above lost the last device events
    # those events are readded
    idx_last_occ = [df_devs[df_devs[DEVICE] == d].iloc[-1].name
                    for d in df_devs[DEVICE].unique()]
    last_occ = df_devs.iloc[idx_last_occ, :]
    df = pd.concat([df, last_occ]) \
        .sort_values(by=TIME) \
        .reset_index(drop=True)

    return df


def _generate_signal(signal, dt='250ms'):
    # Convert strings to timedelta
    dt = pd.Timedelta(dt)
    max_len = pd.Timedelta('0s')
    for i in range(len(signal)):
        signal[i] = (signal[i][0], pd.Timedelta(signal[i][1]))
        max_len += signal[i][1]

    nr_bins = np.floor(max_len / dt).astype(int)
    discretized_sig = np.zeros(nr_bins)
    current_step = 0
    for i in range(len(signal)):
        state = signal[i][0]
        steps = np.floor((signal[i][1] / max_len) * nr_bins).astype(int)
        discretized_sig[current_step:current_step + steps] = 1 if state else -1
        current_step += steps

    # TODO refactor, why does the extrapolation of states do not work out???
    # Eliminate Zeropadding
    last_state = discretized_sig[0]
    for i in range(1, len(discretized_sig)):
        if discretized_sig[i] == 0:
            discretized_sig[i] = last_state
        last_state = discretized_sig[i] 

    return discretized_sig



def create_sig_and_corr(df: pd.DataFrame, hit_idx: int, dt_prae_hit: pd.Timedelta, dt_post_hit: pd.Timedelta,
                        ss_discrete: np.ndarray):
    """ 

    Parameters
    ----------
    df : pd.DataFrame
        

    """
    idx, first, last = get_idxs(df, hit_idx, dt_prae_hit, dt_post_hit)
    tmp = df.loc[idx, [VALUE, 'diff']].values
    tmp = np.insert(tmp, 0, [[not tmp[0, 0], first]], axis=0)
    tmp = np.append(tmp, [[not tmp[-1, 0], last]], axis=0)
    signal = _generate_signal(tmp, dt='250ms')
    corr = sc_signal.correlate(signal, ss_discrete, mode='full')
    return signal, corr


def get_idxs(df, hit_idx, dt_prae_hit, dt_post_hit):
    """ Gets device indices of events before and after to the hit 
        to later on process in a cross correlation.

    Parameters
    ----------
    df : pd.DataFrame
        device dataframe of the selected device
    hit_idx: int
        Indice of the hit
    dt_prae_hit: pd.Timedelta
        How much time to consider prea hit
    dt_prae_hit: pd.Timedelta
        How much time to consider post hit

    Returns
    -------
    list, 
        The involved indices for the signal
    """
    get_df_idx = lambda x: df[df['index'] == x].index.values[0]

    res_idxs = [hit_idx]
    df = df.reset_index()

    if get_df_idx(hit_idx) == 0:
        first = dt_prae_hit
    else:
        # Iterate the events backward and finish if iterated time difference
        # is higher than the origs. signals 
        idx_lower = get_df_idx(hit_idx) - 1
        dt_iter = dt_prae_hit
        dt_prev_tmp = dt_iter - df.at[idx_lower, 'diff']
        while dt_prev_tmp > pd.Timedelta('0s') and idx_lower >= 0:
            dt_iter -= df.at[idx_lower, 'diff']
            res_idxs.insert(0, df.at[idx_lower, 'index'])
            idx_lower -= 1
            dt_prev_tmp -= df.at[idx_lower, 'diff']
        first = dt_iter

    if get_df_idx(hit_idx) == len(df) - 1:
        last = dt_post_hit
    else:
        # Iterate the events forward and finish if iterated time difference
        # is higher than the origs. signals 
        idx_upper = get_df_idx(hit_idx) + 1
        dt_iter = dt_post_hit
        dt_next_tmp = dt_iter - df.at[idx_upper, 'diff']
        while dt_next_tmp > pd.Timedelta('0s') and idx_upper < len(df) - 1:
            dt_iter -= df.at[idx_upper, 'diff']
            res_idxs.append(df.at[idx_upper, 'index'])
            idx_upper += 1
            dt_next_tmp -= df.at[idx_upper, 'diff']
        last = dt_iter

    df = df.set_index('index')
    return res_idxs, first, last



def device_remove_state_matching_signal(df_devs: pd.DataFrame,
                                        #signal: list[tuple[bool, str], ... ],
                                        device: str,
                                        signal: list,
                                        matching_state: int,
                                        corr_threshold=0.2, 
                                        eps_state: str = '0.2s') -> pd.DataFrame:
    """
    Removes states from a device dataframe that match a signal.

    Parameters
    ----------
    df_devs : pd.DataFrame
        A device dataframe
    state_indicator : int
        A number from 0 to length of signal -1 that indicates the state that is going
        to be removed when the signal matches
    eps_state : str, default='0.2s'
        The tolerance a target state may deviate from the given duration in the signal.
    eps_corr : float, default=0.1
        Determines the allowed deviation for the maximium correlation to the correlation of a perfect
        matching signal in percent.

    Example
    -------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> from pyadlml.dataset.devices import device_remove_state_matching_signal
    >>> data = fetch_amsterdam()
    >>> df = data.df_devices.copy()
    >>> sig_original = [ \
            (True, '5s'), \
            (False, '4s'), \
            (True, '1s'), \
            (False, '5s') \
        ] \
    >>> df = device_remove_state_matching_signal(df, sig_original, eps_corr=2)
    """


    matching_state_idx = matching_state
    eps_state = '0.2s'
    # Tolerance for cross correlation w.r.t. auto correlation to highlight as a match
    sig_search = [(s[0], pd.Timedelta(s[1])) for s in signal]


    # Find 
    td = pd.Timedelta(sig_search[matching_state_idx][1])
    state = sig_search[matching_state_idx][0]
    eps_state = pd.Timedelta(eps_state)

    from scipy import signal as sc_signal
    # Create auto correlation and retrieve a proper threshold
    ss_discrete = _generate_signal(sig_search, dt='250ms')
    auto_corr = sc_signal.correlate(ss_discrete, ss_discrete, mode='full')
    auto_corr_max = auto_corr.max()

    # Compute the number of counts that the signal may deviate
    # but still counts as a match
    # For example if a window is 72 units long and the maximum
    # correlation would also be 72 for a signal with the same length. Therefore
    # the eps_corr count would be 14 if the signal is allowed to differ 20%
    if isinstance(corr_threshold, float):
        corr_threshold = int(len(ss_discrete)*corr_threshold)
        corr_threshold = auto_corr_max - corr_threshold

    # Count total length and length up and after matching state of signal
    ss_total_dt, ss_dt_prae_match, ss_dt_post_match = [pd.Timedelta('0s')]*3
    for i, (s, dt) in enumerate(sig_search):
        ss_total_dt += dt
        if i < matching_state_idx:
            ss_dt_prae_match += dt
        if i > matching_state_idx:
            ss_dt_post_match += dt

    plot_dt_around_sig = ss_total_dt*0.25



    def create_hit_list(df_devs):
        df = df_devs[df_devs[DEVICE] == device].copy()
        df['to_convert'] = False
        df['diff'] = pd.Timedelta('0ns')
        df['target'] = False

        df['diff'] = df[TIME].shift(-1) - df[TIME]
        df['target'] = (td - eps_state < df['diff'])\
                    & (df['diff'] < td + eps_state)\
                    & (df[VALUE] == state)

        # Correct the case where the first occurence is already a match
        df.at[df.index[-1], 'diff'] = ss_total_dt

        # Get indices of hits and select first match for display
        hits = df[(df['target'] == True)].index.to_list()
        return hits, df

    hits, df = create_hit_list(df_devs)

    df_sel_dev = df_devs.copy()[df_devs[DEVICE] == device].reset_index()

    for h in hits:
        sig, corr = create_sig_and_corr(df, h, ss_dt_prae_match, ss_dt_post_match, ss_discrete)

        if corr.max() > corr_threshold:
            # Get succeding dev index
            mask = (df_sel_dev['index'] == h)
            idxs_to_drop = df_sel_dev[mask | mask.shift(1)]['index'].values.tolist()
            df_devs = df_devs.drop(idxs_to_drop)

    return df_devs.reset_index(drop=True)




    # Old

    from scipy import signal as sc_signal
    sig_original = signal
    td = pd.Timedelta(sig_original[matching_state][1])
    state = sig_original[matching_state][0]
    eps_state = pd.Timedelta(eps_state)

    # Create perfect correlation yielding a proper threshold
    win = _generate_signal(sig_original, dt='250ms')
    perfect_corr = sc_signal.correlate(win, win, mode='full')
    perfect_corr_max = perfect_corr.max()

    # Compute the number of counts that the signal may deviate
    # but still counts as a match
    # For example if a window is 72 units long and the maximum
    # correlation would also be 72 for a signal with the same length. Therefore
    # the eps_corr count would be 14 if the signal is allowed to differ 20%
    eps_corr = int(len(win)*eps_corr)

    # Create a timelength to append to
    tmp1 = [pd.Timedelta(s[1]) for s in sig_original]
    max_sig = sum(tmp1, pd.Timedelta('0s'))
    states_to_past = [pd.Timedelta(sig_original[s][1]) for s in range(0, matching_state)]
    dt_states_to_past = sum(states_to_past, pd.Timedelta('0s'))
    states_to_future = [pd.Timedelta(sig_original[s][1]) for s in range(matching_state+1, len(sig_original))]
    dt_states_to_future = sum(states_to_future, pd.Timedelta('0s'))

    df = df.copy().reset_index(drop=True)
    df['to_convert'] = False
    df['diff'] = pd.Timedelta('0ns')
    df['target'] = False

    for dev in df[DEVICE].unique():
        dev_mask = (df[DEVICE] == dev)
        max_idx = df.loc[dev_mask].index[-1]
        df.loc[dev_mask, 'diff'] = df.loc[dev_mask, TIME].shift(-1) - df.loc[dev_mask, TIME]
        df.loc[dev_mask, 'target'] = (td - eps_state < df.loc[dev_mask, 'diff']) \
                                     & (df.loc[dev_mask, 'diff'] < td + eps_state) \
                                     & (df.loc[dev_mask, VALUE] == state)
        df.at[max_idx, 'diff'] = max_sig

    for dev in df[DEVICE].unique():
        dev_mask = (df[DEVICE] == dev)
        hits = df[(df['target'] == True) & dev_mask].index.to_list()
        df_dev = df[dev_mask].copy().reset_index()
        for j, h in enumerate(hits):
            idx = [h]
            get_df_idx = lambda i: df_dev.at[i, 'index']

            # Retrieve index of preceeding and succeeding event of the same device
            dev_idx = df_dev[(df_dev['index'] == h)].index[0]
            if dev_idx == 0:
                first = dt_states_to_past
            else:
                dev_lower = dev_idx - 1
                dt_tmp = dt_states_to_past
                dt_prev_tmp = dt_tmp - df_dev.at[dev_lower, 'diff']
                while dt_prev_tmp > pd.Timedelta('0s') and dev_lower >= 0:
                    dt_tmp -= df_dev.at[dev_lower, 'diff']
                    idx.insert(0, get_df_idx(dev_lower))
                    dev_lower -= 1
                    dt_prev_tmp -= df_dev.at[dev_lower, 'diff']
                first = dt_tmp
            if dev_idx == len(df_dev) - 1:
                last = dt_states_to_future
            else:
                dev_upper = dev_idx + 1
                dt_tmp = dt_states_to_future
                dt_next_tmp = dt_tmp - df_dev.at[dev_upper, 'diff']
                while dt_next_tmp > pd.Timedelta('0s') and dev_upper < len(df_dev) - 1:
                    dt_tmp -= df_dev.at[dev_upper, 'diff']
                    idx.append(get_df_idx(dev_upper))
                    dev_upper += 1
                    dt_next_tmp -= df_dev.at[dev_upper, 'diff']
                last = dt_tmp


            # Aasdf
            tmp = df.loc[idx, [VALUE, 'diff']].values
            tmp = np.insert(tmp, 0, [[not tmp[0, 0], first]], axis=0)
            tmp = np.append(tmp, [[not tmp[-1, 0], last]], axis=0)
            signal = _generate_signal(tmp, dt='250ms')

            corr = sc_signal.correlate(signal, win, mode='full')


            is_match = perfect_corr_max - corr.max() < eps_corr
            if is_match:
                df.at[h, 'to_convert'] = True
            #if not is_match:
            #   is_match
            #print(f'match: {is_match}')

    df = df[df['to_convert'] == False].copy()
    df = df[[TIME, DEVICE, VALUE]]
    res = correct_on_off_inconsistency(df)
    return res



import pandas as pd
import numpy as np
from pyadlml.constants import ACTIVITY

ON_APPENDIX = " On"
OFF_APPENDIX = " Off"
def contingency_table_01(X, y):
    """ creates a contingency table for the matrix X and labels y
    Parameters
    ----------
    X : pd.Series or DataFrame
        matrix of 0-1 vectors as rows representing a state of 
        the smart home
    y : pd.Series
        labels for X
    
    Returns
    ---------
    pd.DataFrame
    """
    assert len(X) == len(y), 'Shape mismatch'
    assert isinstance(y, pd.Series)

    X = X.copy().reset_index(drop=True)
    y = y.copy().reset_index(drop=True)

    first_dev = X.columns[0]
    index = X[first_dev]
    res = pd.crosstab(index=index, columns=y)
    res.index = _set_onoffappendix(res.index, first_dev)

    for dev_name in X.columns[1:]:
        tmp = pd.crosstab(index=X[dev_name], columns=y)
        tmp.index = _set_onoffappendix(tmp.index, dev_name)
        res = pd.concat([res, tmp])
    return res

def _set_onoffappendix(index, dev_name):
    """ Using a dataframe index with [true, false] to create a new index with [dev_name + 'on', ...]
    """
    if not index.values[0] and index.values[1]:
        return [dev_name + OFF_APPENDIX, dev_name + ON_APPENDIX]
    elif index.values[0] and not index.values[1]:
        return [dev_name + ON_APPENDIX, dev_name + OFF_APPENDIX]

def cross_correlation(rep):
    """ compute the crosscorelation by comparing for every binary state
    of the smarthome between the devices
    
    Parameters
    ----------
        rep: pd.DataFrame
            device of either raw, lastfired or changepoint representation
            01 vectors of rows representing a state in the smart home
    returns
    -------
        pd.DataFrame (k x k)
        crosscorrelation between each device
    """
    
    dev_lst = rep.columns
    rep = rep.copy().sort_values(by='time')
    rep = rep.where((rep == 1), -1)
    rep = rep.astype('object')
    
    # for every field in a row multiply the each item with the state of the row
    # yields 1 for similar states w.r.t to other devices and -1 for dissimlar
    # values
    def func(series):
        state = series.copy().values
        for i, val in enumerate(state):
            series.iloc[i] = val*state        
        return series
    
    tmp = rep.apply(func, axis=1)
    
    # create pandas dataframe
    arr2d = np.stack(tmp.sum(axis=0).values)
    idx = tmp.columns
    crosstab = pd.DataFrame(data=arr2d, index = tmp.columns, columns=tmp.columns)
    
    # normalize
    crosstab = crosstab/crosstab.iloc[0,0]
    return crosstab


import numpy as np
import pandas as pd
from pyadlml.constants import START_TIME, END_TIME, TIME, NAME, VALUE, DEVICE, NUM, CAT, BOOL
from pyadlml.dataset.stats.util import df_density_binned
from pyadlml.dataset.util import time2int, timestr_2_timedeltas, infer_dtypes, categorical_2_binary
from pyadlml.dataset._core.devices import device_events_to_states, _create_devices, \
    is_device_df, _is_dev_rep2, contains_non_binary, split_devices_binary, create_device_info_dict

from pyadlml.dataset.util import timestr_2_timedeltas
from pyadlml.dataset._representations.raw import create_raw
from pyadlml.util import get_npartitions
from dask import delayed

def device_order_by_count(df_devices):
    return event_count(df_devices)\
        .sort_values(by='event_count', ascending=False)[DEVICE].tolist()


def state_cross_correlation(df_devices,  n_jobs=-1):
    """
    Compute the similarity between devices by comparing the binary values
    for every interval.

    Parameters
    ----------
    df_devices : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    n_jobs : int, default=-1


    Examples
    --------
    >>> from pyadlml.stats import device_state_cross_correlation
    >>> device_state_cross_correlation(data.df_devices)
    device              Cups cupboard  Dishwasher  ...  Washingmachine
    device                                         ...
    Cups cupboard            1.000000    0.997571  ...        0.999083
    Dishwasher               0.997571    1.000000  ...        0.996842
    ...
    Washingmachine           0.999083    0.996842  ...        1.000000
    [14 rows x 14 columns]

    Returns
    -------
    df : pd.DataFrame
        A dataframe of every device against another device. The values range from -1 to 1
        where higher values represent more similarity.
    """
    td = 'td'

    df_devs = df_devices.copy()
    dtypes = infer_dtypes(df_devs)

    # drop numerical devices
    df_devs = df_devs[~df_devs[DEVICE].isin(dtypes[NUM])]
    if dtypes[CAT]:
        df_devs = categorical_2_binary(df_devs, dtypes[CAT])

    def func(row):
        """ gets two rows and returns a crosstab
        """
        try:
            td = row.td.to_timedelta64()
        except:
            return None
        states = row.iloc[1:len(row)-1].values.astype(int)

        for j in range(len(states)):
            row.iloc[1+j] = states[j]*states*td
        return row

    def create_meta(raw):
        return {TIME: 'datetime64[ns]',
                **{name: object for name in raw.columns[1:-1]},
                td: 'timedelta64[ns]'}

    dev_lst = df_devs[DEVICE].unique()
    df_devs = df_devs.sort_values(by=TIME)
    info_dict = create_device_info_dict(df_devs)
    # make off to -1 and on to 1 and then calculate cross correlation between signals
    raw = create_raw(df_devs, info_dict)
    raw.loc[:, dev_lst] = raw.loc[:, dev_lst].applymap(lambda x: 1 if x else -1)
    raw[td] = raw[TIME].shift(-1) - raw[TIME]
    n = get_npartitions()

    # TODO apply leads to to COPY on slice warnings. check out why this is happenign
    with pd.option_context('mode.chained_assignment', None):

        import dask.dataframe as dd
        df = dd.from_pandas(raw.copy(), npartitions=n)\
               .apply(func, axis=1, meta=create_meta(raw))\
               .drop(columns=[TIME, td])\
               .sum(axis=0)\
               .compute(scheduler='threads')

    res = pd.DataFrame(data=np.vstack(df.values), columns=df.index, index=df.index)
    # normalize
    res = res/res.iloc[0, 0]

    #if lst_devs is not None:
    #    for dev in set(lst_devs).difference(set(list(res.index))):
    #        res[dev] = pd.NA
    #        res = res.append(pd.DataFrame(data=pd.NA, columns=res.columns, index=[dev]))
    return res


def state_times(df_devices: pd.DataFrame, binary_state: str = 'on', categorical: bool = True) -> pd.DataFrame:
    """
    Compute times a device is in a certain state.

    Parameters
    ----------
    df_devices : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    binary_state : str one of {'on', 'off'},  default='on'
        pass
    categorical : bool, default=True
        If set, categorical devices are also considered

    Examples
    --------
    >>> from pyadlml.stats import device_on_time
    >>> device_on_time(data.df_devs)
           time          device              td
    0                    Hall-Bedroom door 0 days 00:02:43
    1                    Hall-Bedroom door 0 days 00:00:01
    ...                  ...             ...
    1309                 Frontdoor 0 days 00:00:01
    [1310 rows x 2 columns]

    Returns
    -------
    pd.DataFrame
        A dataframe with two columns, the devices and the time differences.
    """
    assert binary_state in ['on', 'off']
    td = 'td'
    df = df_devices.copy()
    dtypes = infer_dtypes(df)

    # drop numerical devices
    df = df[~df[DEVICE].isin(dtypes[NUM])]

    if categorical and dtypes[CAT]:
        df = categorical_2_binary(df, dtypes[CAT])
    elif not categorical and dtypes[CAT]:
        df = df[~df[DEVICE].isin(dtypes[CAT])]

    # invert binary devices values if the 'off' state is desired
    if binary_state == 'off':
        mask_true = (df[DEVICE].isin(dtypes[BOOL]) & df[VALUE] == True)
        mask_false = (df[DEVICE].isin(dtypes[BOOL]) & df[VALUE] == False)
        df.loc[mask_true, VALUE] = False
        df.loc[mask_false, VALUE] = True

    df = df.sort_values(by=TIME)
    df[td] = pd.NaT
    last_ts = df[TIME].iloc[-1]
    for dev in df[DEVICE].unique():
        diff = df.loc[df[DEVICE] == dev, TIME].shift(-1).copy() \
               - df.loc[df[DEVICE] == dev, TIME].copy()
        diff.iloc[-1] = last_ts - df.loc[df[DEVICE] == dev, TIME].iloc[-1]
        df.loc[(df[DEVICE] == dev), 'td'] = diff
    df[td] = pd.to_timedelta(df[td])
    return df.iloc[:-1, :].loc[(df[VALUE] == True), [TIME, DEVICE, td]]


def state_fractions(df_devices: pd.DataFrame) -> pd.DataFrame:
    """ Computes the fraction a device is in a certain state.
    Categorical devices

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe

    Examples
    --------
    ..code python:
        >>> from pyadlml.stats import state_fractions
        >>> state_fractions(data.df_devices)
                    device                     td_on  ...   frac_on  frac_off
        0           bool_1 0 days 00:00:20.039999994  ...  0.334056  0.665944
        1           bool_2 0 days 00:00:39.999999995  ...  0.666778  0.333222
        ...
        5  cat_2:sub_cat_1 0 days 00:00:09.970000001  ...  0.166194  0.833806
        6  cat_2:sub_cat_2 0 days 00:00:30.020000001  ...  0.500417  0.499583
        7  cat_2:sub_cat_3 0 days 00:00:19.979999992  ...  0.333056  0.666945

    Returns
    -------
    pd.DataFrame
        The fraction a device is in a certain state.

    """
    from pyadlml.dataset.util import infer_dtypes, categorical_2_binary
    df_devs = df_devices.copy()

    dtypes = infer_dtypes(df_devs)
    diff = 'diff'
    td = 'td'
    frac = 'frac'

    # drop numerical devices
    df_devs = df_devs[~df_devs[DEVICE].isin(dtypes[NUM])]

    df_devs = device_events_to_states(df_devs, extrapolate_states=True)
    df_devs = df_devs.sort_values(START_TIME)

    # calculate total time interval for normalization
    int_start = df_devs.iloc[0, 0]
    int_end = df_devs.iloc[df_devs.shape[0] - 1, 1]
    norm = int_end - int_start

    # calculate time deltas for online time
    df_devs[diff] = df_devs[END_TIME] - df_devs[START_TIME]
    df_devs = df_devs.groupby(by=[DEVICE, VALUE])[diff].sum()
    df_devs = pd.DataFrame(df_devs)
    df_devs.columns = [td]

    # compute percentage
    df_devs[frac] = df_devs[td].dt.total_seconds()/norm.total_seconds()

    # TODO refactor, delete flag
    #if lst_devs is not None:
    #    for dev in set(lst_devs).difference(set(list(df_devs.index))):
    #        df_devs = df_devs.append(pd.DataFrame(data=[[pd.NaT, pd.NaT, pd.NA, pd.NA]], columns=df_devs.columns, index=[dev]))
    return df_devs.reset_index()\
        .rename(columns={'index': DEVICE})\
        .sort_values(by=[DEVICE])


def event_count(df_devices: pd.DataFrame) -> pd.DataFrame:
    """ Compute the amount a device is triggered throughout a dataset.

    Parameters
    ----------
    df_devices : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.

    Examples
    --------
    >>> from pyadlml.stats import device_event_count
    >>> device_event_count(data.df_devices)
                    device    event_count
    0        Cups cupboard             98
    1           Dishwasher             42
    ..                 ...            ...
    13      Washingmachine             34

    Returns
    -------
    pd.DataFrame
        The devices and their respective triggercounts.
    """
    assert is_device_df(df_devices)

    col_label = 'event_count'

    ser = df_devices.groupby(DEVICE)[DEVICE].count()
    df_devices = pd.DataFrame({DEVICE: ser.index, col_label: ser.values})

    #if lst_devs is not None:
    #    for dev in set(lst_devs).difference(set(list(df_devices[DEVICE]))):
    #        df_devices = df_devices.append(pd.DataFrame(data=[[dev, 0]],
    #                                        columns=df_devices.columns,
    #                                                    index=[len(df_devices)]))
    return df_devices.sort_values(by=DEVICE)


def inter_event_intervals(df_devices: pd.DataFrame) -> np.ndarray:
    """ Compute the time difference between sucessive device events
        in seconds.

    Parameters
    ----------
    df_devices : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.

    Examples
    --------
    >>> from pyadlml.stats import device_iei
    >>> device_iei(data.df_devices)
    array([1.63000e+02, 3.30440e+04, 1.00000e+00, ..., 4.00000e+00,
           1.72412e+05, 1.00000e+00])

    Returns
    -------
    np.ndarray
        Array of time deltas in seconds.
    """

    # Create timediff to the previous event
    diff_seconds = 'ds'
    df_devices = df_devices.copy().sort_values(by=[TIME])

    # Compute the seconds to the next device
    df_devices[diff_seconds] = df_devices[TIME].diff().shift(-1) / pd.Timedelta(seconds=1)
    return df_devices[diff_seconds].values[:-1]


def resampling_imputation_loss(df_devices: pd.DataFrame, dt: str, return_fraction: bool = False):
    """
    Computes the amount of events neglected when resampling the device dataframe with
    a certain bin size (dt)

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe
    dt : str, one of ['m','h','s','ms']
        The resolution at which  the data is resampled.
    return_fraction : boolean, default=False
        Whether the fraction of imputed events is returned or the total number

    Returns
    -------
    float or int
        Either the fraction or the total amount of lost events depending on the `return_fraction` parameter
    """
    df_devs = df_devices.copy()

    # resample and count the number of events that fall into one bin
    df_devs = df_devs.drop(columns=[VALUE])\
        .set_index(TIME)\
        .resample(dt, kind='timestamp')\
        .count()

    # count the time bins that have more than one event
    # and subtract one event that would survive the imputation
    val_imp = df_devs[(df_devs[DEVICE] > 1)] - 1
    nr_events_disregarded = val_imp.sum()

    if return_fraction:
        return nr_events_disregarded/len(df_devices)
    else:
        return nr_events_disregarded


def event_cross_correlogram_slice(df_devs: pd.DataFrame, lst_devs=None, t_window: str ='20s') -> pd.DataFrame:
    """
    Count the prevalence of devices that trigger within the same time frame.

    Parameters
    ----------
    df_devs : pd.DataFrame
        All recorded devices from a dataset. For more information refer to
        :ref:`user guide<device_dataframe>`.
    lst_devs: list of str, optional
        A list of devices that are included in the statistic. The list can be a
        subset of the recorded devices or contain devices that are not recorded.
    t_window : str of {'[1-12]h', '[1-60]s'}, optional
        Size of the time frame for a single window. The window size
        is given in either n seconds 'ns' or n hours 'nh'. Defaults to 20 seconds '20s'.

    Examples
    --------
    >>> from pyadlml.stats import device_trigger_sliding_window
    >>> device_trigger_sliding_window(data.df_devs)
                       Cups cupboard Dishwasher  ...  Washingmachine
    Cups cupboard                332         10  ...               0
    Dishwasher                    10         90  ...
    ...                          ...        ...  ...             ...
    Washingmachine                 0          0  ...              86

    Returns
    -------
    pd.DataFrame
        TODO
    """

    t_window = timestr_2_timedeltas(t_window)[0]
    
    # create timediff to the previous trigger
    df_devs = df_devs.copy()
    df_devs['time_diff'] = df_devs[TIME].diff()

    #knn
    #    do cumsum for row_duration 
    #    for each row mask the rows that fall into the given area
    if lst_devs is not None:
        dev_list = lst_devs
    else:
        dev_list =  df_devs.device.unique()
    
    df_devs.iloc[0, 3] = pd.Timedelta(0, 's')
    df_devs['cum_sum'] = df_devs['time_diff'].cumsum()
    
    # create cross table with zeros
    res_df = pd.DataFrame(columns=dev_list, index=dev_list)
    for col in res_df.columns:
        res_df[col].values[:] = 0

    # this whole iterations can be done in parallel
    for row in df_devs.iterrows():
        td = row[1].cum_sum
        dev_name = row[1].device

        df_devs['tmp'] = (td - t_window < df_devs['cum_sum']) & (df_devs['cum_sum'] < td + t_window)
        tmp = df_devs.groupby(DEVICE)['tmp'].sum()
        res_df.loc[dev_name] += tmp

    return res_df.sort_index(axis=0, ascending=True) \
        .sort_index(axis=1, ascending=True) \
        .replace(pd.NA, 0)


def events_one_day(df_devices: pd.DataFrame, dt: str=None) -> pd.DataFrame:
    """
    Divide a day into time bins and compute how many device triggers fall into
    each bin.

    Parameters
    ----------
    df_devices : pd.DataFrame
        A datasets recorded devices. For more information refer to
        :ref:`user guide<device_dataframe>`.
    dt : str of {'[1,24]h', '[1,60]m'}, default=None
        The resolution or binsize the day is divided into.

    Examples
    --------
    >>> from pyadlml.stats import device_trigger_one_day
    >>> device_trigger_one_day(data.df_devices, dt='1h')
    device    Cups cupboard  Dishwasher   ...  Washingmachine
    time                                  ...
    00:00:00            0.0         0.0   ...             0.0
    01:00:00           16.0         0.0   ...             0.0
    ...
    23:00:00            6.0         8.0   ...             2.0

    Returns
    -------
    pd.DataFrame
        A dataframe where the columns are the devices and the rows bin the day.
    """
    df_devices = df_devices.copy()

    if dt is None:
        raise NotImplementedError
    else:
        return df_density_binned(df_devices, column_str=DEVICE, dt=dt)


def event_cross_correlogram(df_devices: pd.DataFrame, binsize: str = '1s', maxlag: str = '2m') -> pd.DataFrame:
    """  Calculate cross correlogram

      ccg = correlogram(t, assignment, binsize, maxlag) calculates the
      cross- and autocorrelograms for all device pairs.
      For device on y axis, the events of device on x-axis happen around that device 

    Parameters
    ----------
    df_devices :
        device dataframe
    binsize : str, default='1s'
        The size of one bin in the correlogram
    maxlag : str, default='2m'
        The size of the window for which spikes should be considered.

    Returns
    -------
    ccg             computed correlograms   #device x #device x #bins 
    bins            bin times relative to center    #bins x 1
    """

    import dask
    ET = 'event_times'
    df_dev = df_devices.copy()

    devices = df_devices[DEVICE].unique()
    n_devs = len(devices)
    maxlag = pd.Timedelta(maxlag).seconds
    binsize = pd.Timedelta(binsize).seconds

    n_bins = int((maxlag / binsize) * 2) + 1    # add +1 for symmetric histogram

    ccg = np.zeros((n_devs, n_devs, n_bins))
    bins = np.linspace(-maxlag, maxlag, n_bins)

    # get times in milliseconds relative to the first event
    start_time = df_dev[TIME].iloc[0]
    df_dev[ET] = (df_dev[TIME] - start_time).dt.seconds

    def calc_hist(df_dev, ref, tar):
        # select reference and target device
        t_ref = df_dev[df_dev[DEVICE] == ref][ET].values
        t_tar = df_dev[df_dev[DEVICE] == tar][ET].values

        hist_sum = np.zeros(n_bins)

        # Fix event of reference and do histogram of target events around that fixation
        for k, event in enumerate(t_ref):

            # In case of auto-correlogram delete the reference event from the target array
            # otherwise the 0-bin will have additional #len(tar) events 
            t_tar_prime = np.delete(t_tar, k) if ref == tar else t_tar

            # create histograms
            hist, _ = np.histogram(t_tar_prime, bins=n_bins, range=(event - maxlag,  event + maxlag), density=False)
            hist_sum += hist
        return hist_sum

    # make dask delayed computation for each field
    result = []
    for i, ref in enumerate(devices):
        for j, tar in enumerate(devices):
            result.append(calc_hist(df_dev, ref, tar))
            #result.append(dask.delayed(calc_hist)(df_dev, ref, tar))
    #result = dask.compute(result, scheduler='threads')[0]

    # reintegrate into array
    for i, ref in enumerate(devices):
        for j, tar in enumerate(devices):
            ccg[i, j, :] = result[i*len(devices)+j] # (n_bins,) -> (n_bins, dev, dev)

    return ccg, bins


def _calc_hist(df_dev, ref, tar, n_bins, maxlag):

    # Select reference and target device
    t_ref = df_dev.loc[df_dev[DEVICE] == ref, 'event_times'].copy().values
    t_tar = df_dev.loc[df_dev[DEVICE] == tar, 'event_times'].copy().values
    # Duplicate target #ref times on 0-axis
    tar_matrix = np.tile(t_tar, (len(t_ref), 1))                # (TA,) -> (RE, TA)
    # Duplicate ref #target times on 1-axis
    shift = np.repeat(t_ref.reshape(-1, 1), len(t_tar), axis=1) # (RE, ) -> (RE, TA)

    # Shift every target (row) by reference fix point => zero-centering
    # i.e. first target tm[0,:] is shifted by 0, since shift[0,:] = 0 pointwise, loop
    shifted_target = tar_matrix - shift     # (RE, TA) -> (RE, TA)

    # Compute histograms around 0, (RE, TA) -> (RE, n_bins, 2) containing hist and edges
    if True:
        tmp2 = np.apply_along_axis(lambda *x, **y: np.histogram(*x, **y)[0], axis=1, arr=shifted_target, 
                                bins=n_bins, range=(-maxlag, maxlag), density=False)
        hist_sum2 = tmp2.sum(axis=0)
        return hist_sum2

    tmp = np.apply_along_axis(np.histogram, axis=1, arr=shifted_target,
                            bins=n_bins, range=(-maxlag, maxlag), density=False)

    # Count over hists (0)
    hist_sum = tmp[:, 0].sum(axis=0)
    return hist_sum



def event_cross_correlogram2(df_devices, binsize='1s', maxlag='2m', v1=True, use_dask=False):
    """  Calculate cross correlogram

      ccg = correlogram(t, assignment, binsize, maxlag) calculates the
      cross- and autocorrelograms for all pairs of clusters with input

    Parameters
    ----------
    df_devices :
        device dataframe
    binsize : str, default='1s'
        The size of one bin in the correlogram
    maxlag : str, default='2m'
        The size of the window for which spikes should be considered.

    Returns
    -------
    ccg             computed correlograms   #bins x #device x #device
    bins            bin times relative to center    #bins x 1
    """
    ET = 'event_times'
    df_dev = df_devices.copy()

    devices = df_devices[DEVICE].unique()
    n_devs = len(devices)
    maxlag = pd.Timedelta(maxlag).seconds
    binsize = pd.Timedelta(binsize).seconds

    n_bins = int((maxlag / binsize) * 2) + 1    # add +1 for symmetric histogram

    ccg = np.zeros((n_devs, n_devs, n_bins))
    bins = np.linspace(-maxlag, maxlag, n_bins)

    # get times in milliseconds relative to the first event
    start_time = df_dev[TIME].iloc[0]
    df_dev[ET] = (df_dev[TIME] - start_time).dt.seconds
    """
    Normally a correlogram between two devices is computed by fixing one event of device 1, computing 
    the histogram of the device 2 eventrain, zero-centered around the fixed event and repeat the procedure 
    for every event of device 1.
    
    The fast implementation repeats the event stream of device 1. Then subtract for devices 2 event stream
    is computed. By subtracting each event  from devices 1 vectorized event stream  all events are zero centered
    
    
    """


    # make dask delayed computation for each field
    result = []
    import itertools
    import dask
    if use_dask:
        df_dev = dask.delayed(df_dev)

    k = 0
    for i, j in itertools.combinations_with_replacement(range(0, len(devices)), 2):
        #print(f'({i},{j}) -> {k}')
        if use_dask:
            result.append(dask.delayed(_calc_hist)(df_dev, devices[i], devices[j], n_bins, maxlag))
        else:
            result.append(_calc_hist(df_dev, devices[i], devices[j], n_bins, maxlag))
        k+=1

    if use_dask:
        result = dask.compute(*result)

    # reintegrate into array
    for k, (i, j) in enumerate(itertools.combinations_with_replacement(range(0, len(devices)), 2)):
        #print(f'{k} -> ({i}, {j})')
        if i != j:
            ccg[i, j, :] = result[k]
            ccg[j, i, :] = np.flip(result[k])
        else:
            ccg[i, i, :] = result[k]

    return ccg, bins, devices




def event_cross_correlogram3(df_devices, binsize='1s', maxlag='2m', fix=[], to=[], use_dask=False):
    """  Calculate event cross correlogram. 
        Device A event is fixed and the events of another device B are counted happening in the range of
        maxlag around the fixed events of device A. 
        Therefore, the histgoram c_ij displays the events of device i happening around the device j. 
        Thus a bar in the histogram c_ij means x amound of device i events happen before/after an event
        of device j

    Parameters
    ----------
    df_devices :
        device dataframe
    fix : list

    to: list

    binsize : str, default='1s'
        The size of one bin in the correlogram
    maxlag : str, default='2m'
        The size of the window for which spikes should be considered.

    Returns
    -------
    ccg             computed correlograms   #bins x #device x #device
    bins            bin times relative to center    #bins x 1
    """
    ET = 'event_times'
    df_dev = df_devices.copy()

    devices = df_devices[DEVICE].unique()
    n_devs = len(devices)

    if not fix and not to:
        n_rows, n_cols = [n_devs]*2
        rows, cols = [devices]*2
    if fix:
        n_cols, cols = len(fix), fix
    if to:
        n_rows, rows = len(to), to


    maxlag = pd.Timedelta(maxlag).seconds

    binsize = pd.Timedelta(binsize).seconds

    n_bins = int((maxlag / binsize) * 2) + 1    # add +1 for symmetric histogram

    ccg = np.zeros((n_rows, n_cols, n_bins))
    bins = np.linspace(-maxlag, maxlag, n_bins)

    # get times in milliseconds relative to the first event
    start_time = df_dev[TIME].iloc[0]
    df_dev[ET] = (df_dev[TIME] - start_time).dt.seconds
    """
    Normally a correlogram between two devices is computed by fixing one event of device 1, computing 
    the histogram of the device 2 eventrain, zero-centered around the fixed event and repeat the procedure 
    for every event of device 1.
    
    The fast implementation repeats the event stream of device 1. Then subtract for devices 2 event stream
    is computed. By subtracting each event  from devices 1 vectorized event stream  all events are zero centered
    
    
    """
    def combination_with_replacement(l1, l2, r=2):
        lst = []
        hm = set()
        for e1 in l1:
            for e2 in l2:
                p = frozenset([e1, e2])
                if p not in hm:
                    hm.add(p)
                    lst.append((e1, e2))
        return lst
    class CombHM():
        def __init__(self, rows, cols):
            self.iter_lst = []
            self.value_dict = dict()
            k=0
            for e1 in rows:
                for e2 in cols:
                    key = (e1, e2)
                    if not self._has_item(key):
                        self.value_dict[key] = None
                        self.iter_lst.append(key)
                        k+=1

        def _has_item(self, key):
            if key in self.value_dict.keys():
                return True
            elif (key[1], key[0]) in self.value_dict.keys():
                return True
            else:
                return False

        def _is_rev_key(self, key):
            if key in self.value_dict.keys():
                return False
            elif (key[1], key[0]) in self.value_dict.keys():
                return True
            else:
                raise AttributeError

        def __iter__(self): 
            return self.iter_lst.__iter__()

        def __next__(self):
            return self.iter_lst.__next__()

        def __setitem__(self, idxs, value):
            if isinstance(idxs, int):
                idxs = self.iter_lst[idxs]
            else:
                assert len(idxs) == 2
                idxs = (idxs[0], idxs[1])

            self.value_dict[idxs] = value

        def __getitem__(self, idxs):
            """
            For a single index return the order in which the elements were added. For
            multiple indices 
            """
            if isinstance(idxs, int):
                idxs = self.iter_lst[idxs]
            else:
                assert len(idxs) == 2
                idxs = (idxs[0], idxs[1])

            if self._has_item(idxs):
                if self._is_rev_key(idxs):
                    return np.flip(self.value_dict[(idxs[1], idxs[0])])
                else:
                    return self.value_dict[idxs]
            else:
                raise AttributeError


    # make dask delayed computation for each field
    result = []
    import itertools
    import dask
    if use_dask:
        df_dev = dask.delayed(df_dev)

    combis = CombHM(rows, cols)

    #if rows != cols:
    if True:
        for r, c in combis:
            # TODO REMOVE Compute when the column events happen w.r.t. to a fixed row event
            # Compute when the row events happen w.r.t. to a fixed column event
            if use_dask:
                combis[r,c] = dask.delayed(_calc_hist)(df_dev, c, r, n_bins, maxlag)
            else:
                combis[r,c] = _calc_hist(df_dev, c, r, n_bins, maxlag)

        if use_dask:
            keys = list(combis)
            for k, v in zip(keys, dask.compute(*[combis[k] for k in keys])):
                combis[k] = v 

        for r, c in itertools.product(range(len(rows)), range(len(cols))):
            ccg[r, c, :] = combis[rows[r], cols[c]]

        return ccg, bins, rows, cols 

    else:
        k = 0
        for i, j in itertools.combinations_with_replacement(range(0, len(devices)), 2):
            #print(f'({i},{j}) -> {k}')
            if use_dask:
                result.append(dask.delayed(_calc_hist)(df_dev, devices[i], devices[j], n_bins, maxlag))
            else:
                result.append(_calc_hist(df_dev, devices[i], devices[j], n_bins, maxlag))
            k+=1

        if use_dask:
            result = dask.compute(*result)

        # reintegrate into array
        for k, (i, j) in enumerate(itertools.combinations_with_replacement(range(0, len(devices)), 2)):
            #print(f'{k} -> ({i}, {j})')
            if i != j:
                ccg[i, j, :] = result[k]
                ccg[j, i, :] = np.flip(result[k])
            else:
                ccg[i, i, :] = result[k]

    return ccg, bins, rows, cols 



from pathlib import Path
import os
from pyadlml.dataset.io import set_data_home
from pyadlml.constants import PRIMARY_COLOR, SECONDARY_COLOR, CM_SEQ_MAP, CM_DIV_MAP, DATA_HOME_FOLDER_NAME
from pyadlml.util import (
    ENV_PARALLEL,
    set_primary_color,
    set_diverging_color,
    set_secondary_color,
    set_sequential_color
)

# Initialize and create data home folder
DATA_HOME = Path('/tmp/').joinpath(DATA_HOME_FOLDER_NAME)
set_data_home(DATA_HOME)

# default for parallel execution
os.environ[ENV_PARALLEL] = str(False)

set_primary_color(PRIMARY_COLOR)
set_secondary_color(SECONDARY_COLOR)
set_diverging_color(CM_DIV_MAP)
set_sequential_color(CM_SEQ_MAP)



import sys
import pathlib

from sklearn.neighbors import KernelDensity
from sklearn.svm import SVC

working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest
from sklearn.ensemble import RandomForestClassifier
from pyadlml.dataset import set_data_home, load_act_assist, TIME
from pyadlml.pipeline import Pipeline
from pyadlml.preprocessing import StateVectorEncoder, LabelMatcher, DropTimeIndex
from pyadlml.model_selection import TimeSeriesSplit, LeaveKDayOutSplit, train_test_split
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from pyadlml.dataset import fetch_amsterdam

SUBJECT_ADMIN_NAME = 'admin'


class TestCrossValidation(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/partial_dataset'
        self.data = load_act_assist(dataset_dir, subjects=[SUBJECT_ADMIN_NAME])
        self.data.df_activities = self.data.df_activities_admin

    def test_ts_split(self):
        df = self.data.df_devices
        cv1 = TimeSeriesSplit(n_splits=3)
        cv2 = TimeSeriesSplit(n_splits=3, return_timestamp=True)

        # compare the two splits (indici based and timestamp based)
        for (train_idx, test_idx), (train_ts, test_ts) in zip(cv1.split(df),
                                                              cv2.split(df)):
            tmp1 = df.iloc[train_idx]
            train_sel = df[(train_ts[0] < df[TIME]) & (df[TIME] < train_ts[1])]
            assert tmp1.equals(train_sel)

            tmp2 = df.iloc[test_idx]
            test_sel = df[(test_ts[0] < df[TIME]) & (df[TIME] < test_ts[1])]
            assert tmp2.equals(test_sel)

        cv3 = TimeSeriesSplit(n_splits=3, temporal_split=True, return_timestamp=False)
        cv4 = TimeSeriesSplit(n_splits=3, temporal_split=True, return_timestamp=True)
        # compare the two splits (indici based and timestamp based)
        for (train_idx, test_idx), (train_ts, test_ts) in zip(cv3.split(df),
                                                              cv4.split(df)):
            tmp1 = df.iloc[train_idx]
            train_sel = df[(train_ts[0] < df[TIME]) & (df[TIME] < train_ts[1])]
            assert tmp1.equals(train_sel)

            tmp2 = df.iloc[test_idx]
            test_sel = df[(test_ts[0] < df[TIME]) & (df[TIME] < test_ts[1])]
            assert tmp2.equals(test_sel)



    def test_lkdo_split(self):
        cv = LeaveKDayOutSplit(k=1, n_splits=2)
        # todo do some asserts

    def test_train_test_split(self):
        x_train, x_test, y_train, y_test = train_test_split(
            self.data.df_devices,
            self.data.df_activities
        )
        # todo do some asserts

        x_train, x_test, y_train, y_test, rnd_day = train_test_split(
            self.data.df_devices,
            self.data.df_activities,
            return_day=True
        )
        # todo do some asserts

if __name__ == '__main__':
    unittest.main()


import pandas as pd 
import numpy as np

def _calc_class_accuracy(y_true=None, y_pred=None):
        """ this is the name used for balanced accuracy where the
        average of the accuracys of all respective classes is taken. This is used
        to counter imbalanced datasets.
        Returns
        -------
        float
            the accuracy
        array k dim
            accuracy of each class
        """
        self._model_metrics = True
        from sklearn.metrics import confusion_matrix
        state_count = self._model.K
        lbls = np.arange(0, state_count)
        # a_ij, observations to be in class i, but are predicted to be in class j
        conf_mat = confusion_matrix(y_true, y_pred, labels=lbls) # type: np.ndarray
        K, _ = conf_mat.shape
        total = conf_mat.sum()
        overall_tpr = conf_mat.diagonal().sum()/total
        class_accs = np.zeros((K), dtype=np.float64)
        for k in range(K):
            kth_col_sum = conf_mat[:,k].sum()
            kth_row_sum = conf_mat[k,:].sum()
            # true alarm, it is true and i say true
            tp = conf_mat[k][k]

            # it is true but i say false
            tn =  kth_row_sum - conf_mat[k][k]

            # false alarm, it is false but i say true
            fp = kth_row_sum - conf_mat[k][k]
            assert fp >= 0 and  tp >= 0 and tn >= 0

            # it is false and i say false
            fn = total - (kth_row_sum + kth_col_sum - conf_mat[k][k])
            class_accs[k] = (tp + tn) / (fp + fn + tn + tp)
        class_acc = class_accs.sum()/K
        return class_acc, class_accs




import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.ticker as ticker

def time2numeric_day(time, res='1ms'):
    """ maps a time to a numeric value for a day 
    """
    diff = time - time.floor('d')
    return int(diff/pd.Timedelta(res))


def create_y(tmp2, t_res='ms'):
    """
    Parameters
    ----------
    tmp2 : pd.DataFrame
        dataframe with timestamp index and one column for the predictions
    t_res: String
        string specifying the sampling rate
    """
    assert t_res in ['ms', 's', 'h', 'min']
    
    
    tmp = tmp2.copy()
    col_name = tmp.columns[0]
    
    # create start and end 
    last_row = pd.Series(data={col_name : tmp.iloc[-1,:][col_name]},
                name=tmp.iloc[-1,:].name.ceil('d'))
    first_row = pd.Series(data={col_name : np.nan},
                name=tmp.iloc[-1,:].name.floor('d'))                              
    tmp = tmp.append(first_row)
    tmp = tmp.append(last_row)
    tmp = tmp.resample(t_res).ffill()
    return tmp

def create_cmap(cmap, n=20):
    assert n <= 20
    new_colors = cmap(np.arange(0, n))
    new_cmap = colors.ListedColormap(new_colors)
    return new_cmap


def plot_true_vs_inf_y(enc_lbl, y_true, y_pred, index=None):
    """
    Parameters
    ----------
    enc_lbl: scikit learn model
    y_true: pd.DataFrame
        time index and activity labels
    y_pred: array like
        activity labels
        
    Returns
    -------
    """
    assert len(y_true) == len(y_pred)
    assert isinstance(y_true, pd.Series)
    assert isinstance(y_pred, np.ndarray)
    
    if isinstance(y_pred[0], str):
        y_pred_vals = enc_lbl.transform(y_pred)
    elif isinstance(y_pred[0], float) or isinstance(y_pred[0], int) \
        or isinstance(y_pred[0], np.int64):
        y_pred_vals = y_pred
    else:
        print(y_pred[0])
        print(type(y_pred[0]))
        raise ValueError
        
    if isinstance(y_true.iloc[0], str):
        y_true_vals = enc_lbl.transform(y_true.values)
    elif isinstance(y_true.iloc[0]):
        y_true_vals = y_true.values
    else:
        raise ValueError
        
    title = 'True vs. inferred labels'
    # Plot the true and inferred discrete states
    n_classes = len(enc_lbl.lbl_enc_.classes_)
    
    if index is not None:
        df_ypred = pd.DataFrame(data=y_pred_vals, index=index, columns=['y_pred'])
        df_ypred = create_y(df_ypred, t_res='s')

        df_ytrue = pd.DataFrame(data=y_true_vals, index=index, columns=['y_true'])
        df_ytrue = create_y(df_ytrue, t_res='s')


        # plot 
        fig, axs = plt.subplots(2, figsize=(15, 5))
        cmap = create_cmap(plt.get_cmap("tab20"), n_classes-1)


        axs[0].imshow(df_ytrue.T.values,
                   aspect="auto",
                   interpolation='none',
                   cmap=cmap,
                   vmin=0,
                   vmax=n_classes-1)
        axs[0].set_ylabel("$y_{\\mathrm{true}}$")
        axs[0].set_yticks([])
        axs[0].set_xticks([])


        im = axs[1].imshow(df_ypred.T,
                        aspect="auto",
                        interpolation='none',
                        cmap=cmap,
                        vmin=0,
                        vmax=n_classes-1)

        axs[1].set_ylabel("$y_{\\mathrm{pred}}$")
        axs[1].set_yticks([])

        # format the colorbar
        def formatter_func(x, pos):
            'The two args are the value and tick position'
            val = val_lookup[x]
            return val

        ticks = np.arange(n_classes)
        val_lookup = dict(zip(ticks, 
                              enc_lbl.inverse_transform(ticks)))

        formatter = plt.FuncFormatter(formatter_func)
        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        fig.colorbar(im, cax=cbar_ax, ticks=ticks, format=formatter)
        im.set_clim(-0.5, n_classes-1.5)


        # format the x-axis
        def func(x,p):
            if True:
                if int(x/k) < 10:
                    return '0{}:00'.format(int(x/k)+1)
                else:
                    return '{}:00'.format(int(x/k)+1)

        # calculate the tick positions for lower image
        a,b = axs[1].get_xlim()
        k = (b-a)/24
        tcks_pos = np.arange(0,24)*k + (-0.5 + k)

        x_locator = ticker.FixedLocator(tcks_pos)
        axs[1].xaxis.set_major_formatter(ticker.FuncFormatter(func))
        axs[1].xaxis.set_major_locator(x_locator)
        axs[1].set_aspect(aspect='auto')
        axs[1].tick_params(labelrotation=45)

        fig.suptitle(title) 
        plt.show()
    else:
         # just plot the sequences after another and ignore time information
        fig, axs = plt.subplots(2, figsize=(15, 5))
        cmap = create_cmap(plt.get_cmap("tab20"), n_classes-1)
        
        y_true_vals = np.expand_dims(y_true_vals, axis=1)
        y_pred_vals = np.expand_dims(y_pred_vals, axis=1)

        axs[0].imshow(y_true_vals.T,
                   aspect="auto",
                   interpolation='none',
                   cmap=cmap,
                   vmin=0,
                   vmax=n_classes-1)
        axs[0].set_ylabel("$y_{\\mathrm{true}}$")
        axs[0].set_yticks([])
        axs[0].set_xticks([])


        im = axs[1].imshow(y_pred_vals.T,
                        aspect="auto",
                        interpolation='none',
                        cmap=cmap,
                        vmin=0,
                        vmax=n_classes-1)

        axs[1].set_ylabel("$y_{\\mathrm{pred}}$")
        axs[1].set_yticks([])

        # format the colorbar
        def formatter_func(x, pos):
            'The two args are the value and tick position'
            val = val_lookup[x]
            return val

        ticks = np.arange(n_classes)
        val_lookup = dict(zip(ticks, 
                              enc_lbl.inverse_transform(ticks)))

        formatter = plt.FuncFormatter(formatter_func)
        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        fig.colorbar(im, cax=cbar_ax, ticks=ticks, format=formatter)
        im.set_clim(-0.5, n_classes-1.5)
        fig.suptitle(title) 
        plt.show()


import json
from dash.dependencies import Output, Input, State
import dash
from dash.exceptions import PreventUpdate
import pandas as pd
from pyadlml.constants import ACTIVITY, DEVICE, END_TIME, START_TIME
from pyadlml.dataset.plot.plotly.activities import bar_count, bar_cum, boxplot_duration, density, heatmap_transitions
from pyadlml.dataset.plot.plotly.acts_and_devs import activity_vs_device_events_hist, contingency_events, contingency_states, event_correlogram
from pyadlml.dataset.plot.plotly.dashboard.layout import BIN_SIZE_SLIDER, LAG_SLIDER
from pyadlml.dataset.stats.activities import activities_dist
from pyadlml.dataset.stats.acts_and_devs import contingency_table_events, contingency_table_states
import plotly.graph_objects as go
from pyadlml.dataset.util import activity_order_by, device_order_by, num_to_timestamp, select_timespan


def bind_toggle_collapse(name):
    def toggle_collapse(n, is_open):
        return bool(n) ^ bool(is_open)

    toggle_collapse.__name__ = name
    return toggle_collapse

def _initialize_activity_toggle_cbs(app, act_id):

    app.callback(
        Output(f"clps-act-transition-{act_id}", "is_open"),
        [Input(f"clps-act-transition-button-{act_id}", "n_clicks")],
        [State(f"clps-act-transition-{act_id}", "is_open")],
    )(bind_toggle_collapse(f'toggle_collapse_act_trans_{act_id}'))

    app.callback(
        Output(f"clps-act-boxplot-{act_id}", "is_open"),
        [Input(f"clps-act-boxplot-button-{act_id}", "n_clicks")],
        [State(f"clps-act-boxplot-{act_id}", "is_open")],
    )(bind_toggle_collapse(f'toggle_collapse_act_bp_{act_id}'))


    app.callback(
        Output(f"clps-act-bar-{act_id}", "is_open"),
        [Input(f"clps-act-bar-button-{act_id}", "n_clicks")],
        [State(f"clps-act-bar-{act_id}", "is_open")],
    )(bind_toggle_collapse(f'toggle_collapse_act_bar_{act_id}'))


    app.callback(
        Output(f"clps-avd-state-{act_id}", "is_open"),
        [Input(f"clps-avd-state-button-{act_id}", "n_clicks")],
        [State(f"clps-avd-state-{act_id}", "is_open")],
    )(bind_toggle_collapse(f'toggle_collapse_avd_state_{act_id}'))

    app.callback(
        Output(f"clps-avd-event-{act_id}", "is_open"),
        [Input(f"clps-avd-event-button-{act_id}", "n_clicks")],
        [State(f"clps-avd-event-{act_id}", "is_open")],
    )(bind_toggle_collapse(f'toggle_collapse_avd_event_{act_id}'))


    app.callback(
        Output(f"clps-avd-hist-{act_id}", "is_open"),
        [Input(f"clps-avd-hist-button-{act_id}", "n_clicks")],
        [State(f"clps-avd-hist-{act_id}", "is_open")],
    )(bind_toggle_collapse(f'toggle_collapse_avd_hist_{act_id}'))

    app.callback(
        Output(f"clps-avd-cc-{act_id}", "is_open"),
        [Input(f"clps-avd-cc-button-{act_id}", "n_clicks")],
        [State(f"clps-avd-cc-{act_id}", "is_open")],
    )(bind_toggle_collapse(f'toggle_collapse_avd_cc_{act_id}'))

def _create_activity_tab_callback(app, act_id, df_acts, start_time, end_time, plt_height_acts):

    # TODO Circual imports hack
    from pyadlml.dataset.plot.plotly.util import dash_get_trigger_value
    from pyadlml.dataset.plot.plotly.dashboard.dashboard import _is_trigger

    @app.callback(
        output=[
            Output(f'acts_graph-bar-{act_id}', 'figure'),
            Output(f'acts_graph-boxplot-{act_id}', 'figure'),
            Output(f'acts_graph-density-{act_id}', 'figure'),
            Output(f'acts_graph-transition-{act_id}', 'figure'),
            Output(f'acts_activity-order-{act_id}', 'data'),
            Output(f'acts_density-data-{act_id}', 'data')
        ],
        inputs=[
            Input('tabs', 'active_tab'),
            Input(f'act-trigger-{act_id}', 'children'),
            Input('range-slider', 'value'),
            Input('select-activities', 'value'),
            Input(f'acts_bp-drop-{act_id}', 'value'),
            Input(f'acts_bar-drop-{act_id}', 'value'),
            Input(f'acts_bp-scale-{act_id}', 'value'),
            Input(f'acts_bar-scale-{act_id}', 'value'),
            Input(f'acts_trans-scale-{act_id}', 'value'),
            Input(f'acts_sort-{act_id}', 'value'),
        ],
        state=[
            State(f'acts_graph-bar-{act_id}', 'figure'),
            State(f'acts_graph-boxplot-{act_id}', 'figure'),
            State(f'acts_graph-density-{act_id}', 'figure'),
            State(f'acts_graph-transition-{act_id}', 'figure'),
            State(f'acts_activity-order-{act_id}', 'data'),
            State(f'acts_density-data-{act_id}', 'data'),
        ]

    )
    def update_activity_tab(active_tab, act_trigger, rng, sel_activities,
                               drop_box: str, drop_bar: str, scale_boxplot: str,
                               scale_bar: str, scale_trans: str, act_order_trigger,
                               fig_bar, fig_bp, fig_dens, fig_trans, act_order, act_density,
        ):

        ctx = dash.callback_context
        if dash_get_trigger_value(ctx) is None or active_tab != f'tab-acts-{act_id}':
            raise PreventUpdate

        try:
            act_order = json.loads(act_order)
        except TypeError:
            # happens if no act-order is initialized
            act_order = None
        try:
            act_density = pd.read_json(act_density)
        except:
            act_density = None

        # Filter selected timeframe, activities and devices
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_acts = select_timespan(df_acts=df_acts, start_time=st, end_time=et,
                                       clip_activities=True)
        curr_df_acts = curr_df_acts[curr_df_acts[ACTIVITY].isin(sel_activities)]
        # TODO refactor, where is the y_label column coming from
        curr_df_acts = curr_df_acts[[START_TIME, END_TIME, ACTIVITY]]

        # Get update type
        is_trigger_bar_drop = _is_trigger(ctx, f'acts_bar-drop-{act_id}')
        is_trigger_bar_scale = _is_trigger(ctx, f'acts_bar-scale-{act_id}')
        is_trigger_sort = _is_trigger(ctx, f'acts_sort-{act_id}')
        is_trigger_range = _is_trigger(ctx, 'range-slider')
        is_trigger_bp_drop = _is_trigger(ctx, f'acts_bp-drop-{act_id}')
        is_trigger_bp_scale = _is_trigger(ctx, f'acts_bp-scale-{act_id}')
        is_trigger_trans_scale = _is_trigger(ctx, f'acts_trans-scale-{act_id}')

        data_update = is_trigger_range or _is_trigger(ctx, 'select-activities')
        # Determine the intent, when the trigger was the avd plot
        signal_reset_bp = _is_trigger(ctx, f'act-trigger-{act_id}') and act_trigger == 'reset_sel_bp'
        signal_reset_trans = _is_trigger(ctx, f'act-trigger-{act_id}') and act_trigger == 'reset_sel_trans'
        signal_reset_all = _is_trigger(ctx, f'act-trigger-{act_id}') and act_trigger == 'reset_sel'

        order_update = ((is_trigger_sort or (is_trigger_bar_drop and act_order_trigger == 'value')) \
                       and not signal_reset_all) \
                       or act_order is None

        bp_update = is_trigger_bp_drop or is_trigger_bp_scale \
                    or data_update or order_update or signal_reset_all\
                    or signal_reset_bp
        bar_update = is_trigger_bar_drop or order_update or data_update \
                     or is_trigger_bar_scale
        trans_update = data_update or order_update or is_trigger_trans_scale \
                       or signal_reset_all

        # If the activity-order is changed or the bar plot is changed
        # and would change the order
        if order_update:
            if act_order_trigger == 'value' or is_trigger_bar_drop:
                act_order_trigger = 'duration' if drop_bar == 'cum' else 'count'
            act_order = activity_order_by(curr_df_acts, act_order_trigger)

        # Update activity bars
        if bar_update:
            if drop_bar == 'count':
                fig_bar = bar_count(curr_df_acts, order=act_order, scale=scale_bar,
                                    height=plt_height_acts)
            else:
                fig_bar = bar_cum(curr_df_acts, order=act_order, scale=scale_bar,
                                  height=plt_height_acts)

        # Update log for boxplot
        if bp_update:
            #if _get_trigger_value(ctx) == 'vp':
            #    fig_bp = violin_duration(curr_df_acts, order=act_order, scale=scale_boxplot)
            #else:
            #    fig_bp = boxplot_duration(curr_df_acts, order=act_order, scale=scale_boxplot)
            fig_bp = boxplot_duration(curr_df_acts, order=act_order, scale=scale_boxplot,
                                      height=plt_height_acts)

        # Only update the act_density matrix if it is
        if data_update or act_density is None:
            act_density = activities_dist(curr_df_acts.copy(), n=1000, dt=None)
        if order_update or data_update:
            fig_dens = density(df_density=act_density, order=act_order, height=plt_height_acts)

        if trans_update:
            fig_trans = heatmap_transitions(curr_df_acts, order=act_order, scale=scale_trans,
                                            height=plt_height_acts)

        return fig_bar, fig_bp, fig_dens, fig_trans, json.dumps(list(act_order)), act_density.to_json()



def _create_acts_vs_devs_tab_callback(app, act_id, df_acts, df_devs, start_time, end_time):

    from pyadlml.dataset.plot.plotly.dashboard.dashboard import _is_trigger
    from pyadlml.dataset.plot.plotly.util import dash_get_trigger_value

    @app.callback(
        output=[
            Output(f'avd_state-contingency-{act_id}', 'data'),
            Output(f'avd_event-contingency-{act_id}', 'data'),
        ],
        inputs=[
            Input(f'avd-trigger-{act_id}', 'children'),
            Input(f'avd-update-{act_id}', 'children'),
        ],
        state=[
            State('range-slider', 'value'),
            State('select-activities', 'value'),
            State('select-devices', 'value'),
            State(f'avd_state-contingency-{act_id}', 'data'),
            State(f'avd_event-contingency-{act_id}', 'data'),
        ]
    )
    def update_contingency_tables(trigger, update, rng, sel_activities, sel_devices, state_cont, event_cont):
        """ Contingency tables take very long to create. Therefore generate data unrelated to current
            dashboard activities.
        
        """

        ctx = dash.callback_context
        
        if dash_get_trigger_value(ctx) is None:
            raise PreventUpdate

        try:
            df_con_states = pd.read_json(state_cont)
            df_con_states = df_con_states.astype('timedelta64[ns]')
            df_con_events = pd.read_json(event_cont)
        except ValueError:
            df_con_states = None
            df_con_events = None

        is_data_update = _is_trigger(ctx, 'new_data')

        # Filter selected timeframe, activities and devices
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_devs, curr_df_acts = select_timespan(df_acts=df_acts, df_devs=df_devs,
                                                     start_time=st, end_time=et, clip_activities=True)
        curr_df_acts = curr_df_acts[curr_df_acts[ACTIVITY].isin(sel_activities)]
        curr_df_devs = curr_df_devs[curr_df_devs[DEVICE].isin(sel_devices)]
        # TODO refactor, where is the y_label column coming from
        curr_df_acts = curr_df_acts[[START_TIME, END_TIME, ACTIVITY]]


        # If data has changed the contingency tables have to be recomputed
        if is_data_update or df_con_states is None or df_con_events is None:
            # Recompute state contingency
            df_con_states = contingency_table_states(curr_df_devs, curr_df_acts,
                                                     n_jobs=4)
            df_con_events = contingency_table_events(curr_df_devs, curr_df_acts)
            update_data = True
        else:
            update_data = False

        if update_data:
            state_dump, event_dump = df_con_states.to_json(date_unit="ns"), df_con_events.to_json()
        else:
            state_dump, event_dump = dash.no_update, dash.no_update

        return state_dump, event_dump

    @app.callback(
        output=[
            Output(f'avd_graph-event-contingency-{act_id}', 'figure'),
            Output(f'avd_graph-state-contingency-{act_id}', 'figure'),
            Output(f'avd_graph-hist-{act_id}', 'figure'),
            Output(f'avd_graph-cc-{act_id}', 'figure'),
            Output(f'loading-output-{act_id}', 'children'),
            Output(f'avd-update-{act_id}', 'children'),
            Output(f'avd_activity-order-{act_id}', 'data'),
            Output(f'avd_device-order-{act_id}', 'data'),
        ],
        inputs=[
            Input('tabs', 'active_tab'),
            Input(f'avd-trigger-{act_id}', 'children'),
            Input(f'avd-update-{act_id}', 'children'),
            Input(f'avd_state-scale-{act_id}', 'value'),
            Input(f'avd_event-scale-{act_id}', 'value'),
            Input(f'avd_act-order-trigger-{act_id}', 'value'),
            Input(f'avd_dev-order-trigger-{act_id}', 'value'),
            Input(f'avd_hist-normalize-{act_id}', 'value'),
            Input(f'avd_hist-sel-dev-{act_id}', 'value'),
            Input(f'avd_hist-sel-act-{act_id}', 'value'),
            Input(f'avd_cc-sel-fix-{act_id}', 'value'),
            Input(f'avd_cc-sel-to-{act_id}', 'value'),
            Input(f'avd_cc-lag-slider-{act_id}', 'value'),
            Input(f'avd_cc-binsize-slider-{act_id}', 'value'),
        ],
        state=[
            State('range-slider', 'value'),
            State('select-activities', 'value'),
            State('select-devices', 'value'),
            State(f'avd_state-contingency-{act_id}', 'data'),
            State(f'avd_event-contingency-{act_id}', 'data'),
            State(f'avd_activity-order-{act_id}', 'data'),
            State(f'avd_device-order-{act_id}', 'data'),
            State(f'avd_graph-hist-{act_id}', 'figure'),
            State(f'avd_graph-cc-{act_id}', 'figure'),
        ]
    )
    def update_acts_vs_devs_tab(active_tab, trigger, update,  state_scale,  event_scale,
                                act_order_trigger, dev_order_trigger, 
                                hist_norm, hist_sel_dev, hist_sel_act,
                                cc_sel_fix, cc_sel_to, cc_lag, cc_binsize,
                                rng, sel_activities, sel_devices,
                                state_cont, event_cont, act_order, dev_order,
                                fig_hist, fig_cc
                                
        ):

        ctx = dash.callback_context

        if dash_get_trigger_value(ctx) is None or active_tab != f'tab-acts_vs_devs-{act_id}':
            raise PreventUpdate

        try:
            df_con_states = pd.read_json(state_cont)
            df_con_states = df_con_states.astype('timedelta64[ns]')
            df_con_events = pd.read_json(event_cont)
        except ValueError:
            df_con_states = None
            df_con_events = None
        except BaseException as err:
            print(f"Unexpected {err=}, {type(err)=}")
            raise

        try:
            act_order = json.loads(act_order)
            dev_order = json.loads(dev_order)
        except TypeError:
            act_order = None
            dev_order = None


        # Filter selected timeframe, activities and devices
        st = num_to_timestamp(rng[0], start_time=start_time, end_time=end_time)
        et = num_to_timestamp(rng[1], start_time=start_time, end_time=end_time)
        curr_df_devs, curr_df_acts = select_timespan(df_acts=df_acts, df_devs=df_devs,
                                                     start_time=st, end_time=et, clip_activities=True)
        curr_df_acts = curr_df_acts[curr_df_acts[ACTIVITY].isin(sel_activities)]
        curr_df_devs = curr_df_devs[curr_df_devs[DEVICE].isin(sel_devices)]
        # TODO refactor, where is the y_label column coming from
        curr_df_acts = curr_df_acts[[START_TIME, END_TIME, ACTIVITY]]

        # If the activity-order is changed or the bar plot is changed
        # and would change the order
        is_data_update = _is_trigger(ctx, 'new_data')
        is_trigger_act_order = _is_trigger(ctx, f'avd_act-order-trigger-{act_id}')
        is_trigger_dev_order = _is_trigger(ctx, f'avd_dev-order-trigger-{act_id}')
        is_trigger_hist_norm = _is_trigger(ctx, f'avd_hist-normalize-{act_id}') 
        is_trigger_hist_sel_dev = _is_trigger(ctx, f'avd_hist-sel-dev-{act_id}')
        is_trigger_hist_sel_act = _is_trigger(ctx, f'avd_hist-sel-act-{act_id}')
        is_trigger_cc_lag = _is_trigger(ctx, f'avd_cc-lag-slider-{act_id}') 
        is_trigger_cc_binsize = _is_trigger(ctx, f'avd_cc-binsize-slider-{act_id}') 
        is_trigger_cc_sel_fix = _is_trigger(ctx, f'avd_cc-sel-fix-{act_id}')
        is_trigger_cc_sel_to = _is_trigger(ctx, f'avd_cc-sel-to-{act_id}')


        order_update = is_trigger_act_order or is_trigger_dev_order\
                       or act_order is None or dev_order is None
        hist_update = is_trigger_hist_norm or is_trigger_hist_sel_dev or is_trigger_hist_sel_act
        cc_update = is_trigger_cc_lag or is_trigger_cc_binsize or is_trigger_cc_sel_fix \
                 or is_trigger_cc_sel_to

        if order_update:
            act_order = activity_order_by(curr_df_acts, act_order_trigger)
            dev_order = device_order_by(curr_df_devs, dev_order_trigger)


        if active_tab != f'tab-acts_vs_devs-{act_id}' and is_data_update:
            return dash.no_update, dash.no_update, dash.no_update, dash.no_update, dash.no_update, 'updated', \
                   df_con_states.to_json(), df_con_events.to_json(), \
                   json.dumps(list(act_order)), json.dumps(list(dev_order))
        elif active_tab != f'tab-acts_vs_devs-{act_id}':
            raise PreventUpdate

        # Create figures
        if hist_update:
            assert hist_norm in ['True', 'False']
            fig_hist = activity_vs_device_events_hist(
                curr_df_devs,
                curr_df_acts,
                device=hist_sel_dev,
                activity=hist_sel_act,
                normalize=eval(hist_norm),
                height=250,
            )

        if cc_update:
            fig_cc = event_correlogram(
                curr_df_devs, 
                curr_df_acts, 
                fix=cc_sel_fix,
                to=cc_sel_to, 
                maxlag=LAG_SLIDER[cc_lag], 
                binsize=BIN_SIZE_SLIDER[cc_binsize],
                use_dask=True
            )

        if df_con_states is not None:
            fig_adec = contingency_events(
                        con_tab=df_con_events, scale=event_scale,
                        act_order=act_order, dev_order=dev_order
            )
        else:
            fig_adec = go.Figure()

        if df_con_events is not None: 
            fig_adsc = contingency_states(
                        df_devs=curr_df_devs,   # for event order
                        con_tab=df_con_states, scale=state_scale,
                        act_order=act_order, dev_order=dev_order
            )
        else:
            fig_adsc = go.Figure()


        return fig_adec, fig_adsc, fig_hist, fig_cc, None, 'needs update',  \
               json.dumps(list(act_order)), json.dumps(list(dev_order))




from setuptools import setup, find_packages

with open('README.md', 'r') as readme_file:
    README = readme_file.read()

with open('HISTORY.md', 'r') as history_file:
    HISTORY = history_file.read()


# Most basic version
install_requires = ['numpy', 'pandas', 'joblib']


_extras_light = ['mega.py', 'dask[complete]',
                 'matplotlib', 'scipy', 'sklearn']

_extras_datavis = [
    'sqlalchemy',
    'plotly',
    'dash_daq',
    'dash-bootstrap-components',
    'dash',
]

_extras_all = _extras_light + _extras_datavis



setup_args = dict(
    name="pyadlml",
    version="0.0.7.15-alpha",
    url="https://github.com/tcsvn/pyadlml",
    author="Christian Meier",
    description="Sklearn flavored library containing numerous Activity of Daily Livings datasets, preprocessing methods, visualizations and models.",
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    author_email="account@meier-lossburg.de",
    license="MIT",
    packages=find_packages(),
    install_requires=install_requires,
    extras_require=dict(
        complete=_extras_all,
        datavis=_extras_datavis,
        light=_extras_light,
    ),
    keywords=['Activity of Daily Living'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Visualization"
    ],
)


if __name__ == '__main__':
    setup(**setup_args)



import unittest
from hassbrain_algorithm.models.hmm.hmm import ModelHMM_log_scaled
from hassbrain_algorithm.controller import Controller
from hassbrain_algorithm.controller import Dataset
from hassbrain_algorithm.datasets.homeassistant import DatasetHomeassistant


class TestHomeassistantModelHMM(unittest.TestCase):

    # Model part

    def setUp(self):
        # set of observations
        self.ctrl = Controller()
        self.ctrl.set_dataset(Dataset.HASS_TESTING)
        self.hass_obj = self.ctrl._dataset #type: DatasetHomeassistant

    def tearDown(self):
        pass

    def test_load_custom_lists_modelHMM(self):
        custom_state_list = ['sleeping', 'cooking']
        custom_obs_list = [
            'binary_sensor.motion_bed',
            'binary_sensor.motion_mirror',
            'binary_sensor.motion_pc',
            'switch.test_switch_1',
            'light.test_light'
        ]
        self.ctrl.set_custom_state_list(custom_state_list)
        self.ctrl.set_custom_obs_list(custom_obs_list)

        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)


    def test_load_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)
        print(self.hass_obj.get_obs_lbl_hashmap())
        print(self.hass_obj.get_state_lbl_hashmap())


    def test_train_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

        self.ctrl.train_model()
        print(self.ctrl._model)


    def test_bench_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

        self.ctrl.register_benchmark()
        self.ctrl.train_model()
        print(self.ctrl._model)
        report = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True
        )
        print(report)

    def test_classify(self):
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('-'*10)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        pred_state = hmm_model.classify(obs_seq)
        print('#'*100)
        print(pred_state)


    def test_classify_multi(self):
        """
        used to test for classification of multiple labels
        """
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('-'*10)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1)]#, ('binary_sensor.motion_bed', 0)]
        act_state_dict = hmm_model.classify_multi(obs_seq)
        print('#'*100)
        #print(act_state_dict)
        #print(hmm_model.get_state_label_list())



    def test_pred_next_obs_single(self):
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        #print(hmm_model)
        print('#'*100)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        tupel = hmm_model.predict_next_obs(obs_seq)

    def test_pred_next_obs_multi(self):
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        #print(hmm_model)
        print('#'*100)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        #arr = hmm_model.predict_next_obs_arr(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        res_dict = hmm_model.predict_prob_xnp1(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        res_dict = hmm_model.predict_prob_xnp1(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        print('#'*100)
        print(res_dict)


    def test_bench_modelLogScaledHMM(self):
        self.ctrl.load_dataset()
        hmm_model = ModelHMM_log_scaled(self.ctrl)
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

        self.ctrl.register_benchmark()
        self.ctrl.train_model()
        print(self.ctrl._model)
        report = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True
        )
        print(report)


class TestHomeassistantModelHMMLogScaled(unittest.TestCase):

    # Model part
    def setUp(self):
        # set of observations
        self.ctrl = Controller()
        self.ctrl.set_dataset(Dataset.HASS_TESTING)
        self.hass_obj = self.ctrl._dataset #type: DatasetHomeassistant
        self.hmm_model = ModelHMM_log_scaled(self.ctrl)

    def tearDown(self):
        pass

    def test_load_custom_lists_modelHMM(self):
        custom_state_list = ['sleeping', 'cooking']
        custom_obs_list = [
            'binary_sensor.motion_bed',
            'binary_sensor.motion_mirror',
            'binary_sensor.motion_pc',
            'switch.test_switch_1',
            'light.test_light'
        ]
        hmm_model = self.hmm_model
        self.ctrl.set_custom_state_list(custom_state_list)
        self.ctrl.set_custom_obs_list(custom_obs_list)

        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

    def test_load_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)
        print(self.hass_obj.get_obs_lbl_hashmap())
        print(self.hass_obj.get_state_lbl_hashmap())

    def test_train_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

        self.ctrl.train_model()
        print(self.ctrl._model)

    def test_bench_modelHMM(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        print(self.ctrl._model)

        self.ctrl.register_benchmark()
        self.ctrl.train_model()
        print(self.ctrl._model)
        report = self.ctrl.create_report(
            conf_matrix=True,
            accuracy=True,
            precision=True,
            recall=True,
            f1=True
        )
        print(report)

    def test_bench_train_loss(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        self.ctrl.register_benchmark()
        #self.ctrl.register_loss_file_path('/home/cmeier/code/tmp/train_loss.log')
        self.ctrl.register_loss_file_path('/home/cmeier/code/data/hassbrain/media/tmp/trainloss.csv')
        self.ctrl.train_model()
        self.ctrl.show_plot()

    def test_bench_train_loss_and_graph(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        self.ctrl.register_benchmark()
        self.ctrl.register_loss_file_path('/home/cmeier/code/tmp/train_loss.csv')
        self.ctrl.train_model()

        self.ctrl.save_loss_plot_to_file('/home/cmeier/code/tmp/train_loss_graph.png')
        #self.ctrl.save_visualization_to_file('/home/cmeier/code/data/tmp/train_loss_graph.img')

    def test_generate_visualization(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        self.ctrl.save_visualization_to_file('/home/cmeier/code/tmp/visualization.png')


    def test_bench_train_loss_and_graph(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        hmm_model._hmm.set_format_full(True)
        self.ctrl.register_benchmark()
        self.ctrl.register_loss_file_path('/home/cmeier/code/tmp/train_loss.csv')
        self.ctrl.train_model()
        self.ctrl.save_loss_plot_to_file('/home/cmeier/code/tmp/train_loss_graph.png')
        self.ctrl.save_visualization_to_file('/home/cmeier/code/tmp/visualization.png')

    def test_classify(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('-'*10)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        pred_state = hmm_model.classify(obs_seq)
        print('#'*100)
        print(pred_state)

    def test_classify_multi(self):
        """
        used to test for classification of multiple labels
        """
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        #self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('-'*10)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1)]#, ('binary_sensor.motion_bed', 0)]
        act_state_dict = hmm_model.classify_multi(obs_seq)
        print('#'*100)
        print(act_state_dict)
        #print(act_state_dict)
        #print(hmm_model.get_state_label_list())

    def test_pred_next_obs_single(self):
        self.ctrl.load_dataset()
        hmm_model = self.hmm_model
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        #print(hmm_model)
        print('#'*100)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        tupel = hmm_model.predict_next_obs(obs_seq)
        print(tupel)

    def test_pred_next_obs_multi(self):
        hmm_model = self.hmm_model
        self.ctrl.load_dataset()
        self.ctrl.register_model(hmm_model)
        self.ctrl.init_model_on_dataset()
        self.ctrl.train_model()
        hmm_model._hmm.set_format_full(True)
        print(hmm_model)
        print('#'*100)
        obs_seq = [('binary_sensor.motion_bed', 0), ('binary_sensor.motion_mirror', 1), ('binary_sensor.motion_bed', 0)]
        #arr = hmm_model.predict_next_obs_arr(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        res_dict = hmm_model.predict_prob_xnp1(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        res_dict = hmm_model.predict_prob_xnp1(obs_seq)
        print(hmm_model._obs_lbl_hashmap)
        print(hmm_model._obs_lbl_rev_hashmap)
        print('#'*100)
        print(res_dict)






from pyadlml.constants import DATA_DCT_KEY_ACTS, DATA_DCT_KEY_DEVS, DEVICE, ACTIVITY
from pyadlml.dataset._core.activities import ActivityDict
from .local import get_data_home, _ensure_dh_folder_exists, _delete_data
from pathlib import Path
import pandas as pd
import joblib
from abc import ABC, abstractmethod




class DataFetcher(ABC):
    FN_CLEANED = 'cleaned%s.joblib'
    FN_CACHED = 'cached%s.joblib'

    def __init__(self, dataset_name: str,
                       downloader,
                       correct_activities=False,
                       correct_devices=False
        ):

        self.downloader = downloader
        self.ds_name = dataset_name
        self.apply_dev_corr = correct_devices
        self.apply_act_corr = correct_activities


        self.data_home = get_data_home()
        self._dataset_folder = None
        self._original_folder = None


    def _create_exp_folder(self):
        """
        """
        _ensure_dh_folder_exists()
        self.data_home = get_data_home()

    @property
    def dataset_folder(self):
        return Path(self.data_home).joinpath(self.ds_name)

    @property
    def original_folder(self):
        return self.dataset_folder.joinpath('original')

    def correct_activities(self, subject, df_activities, ident=None):
        """ May be overriden by child classes
        """
        from pyadlml.dataset._core.activities import correct_activities
        return correct_activities(df_activities, retain_corrections = self.retain_corrections)

    def apply_corrections(self, data:dict, ident:str=None, retain_corrections=False):
        """ Applies corrections

        Parameters
        ----------
        ident : None or str


        data : dict
            Of the form 
            {
                'activity_list': [],
                'device_list': [],
                'devices': pd.DataFrame,
                'activities': pd.DataFrame,
                ...
            }
        """
        if self.apply_act_corr:
            acts = data[DATA_DCT_KEY_ACTS]
            unpack = isinstance(acts, pd.DataFrame)
            if unpack:
                acts = ActivityDict.wrap(acts)

            for key, df in acts.items():
                df_acts, corrections = self.correct_activities(key, df, ident)
                acts[key] = df_acts
                if corrections:
                    if 'correction_activities' not in data.keys():
                        data['correction_activities'] = {}
                    data['correction_activities'][key] = corrections

            if unpack:
                data[DATA_DCT_KEY_ACTS] = df_acts
                if corrections:
                    data['correction_activities'] = data['correction_activities'][list(acts.keys())[0]]

        if self.apply_dev_corr:
            from pyadlml.dataset._core.devices import correct_devices
            df_dev, correction_dev_dict = correct_devices(data[DATA_DCT_KEY_DEVS], 
                                retain_correction=retain_corrections)
            data[DATA_DCT_KEY_DEVS] = df_dev
            if correction_dev_dict:
                data.update(correction_dev_dict)

        return data

    def _gen_filenames_cached(self, identifier):
        """ Creates unique cache names when i.e. more subjecta have to be 
            loaded.
        """
        
        if identifier is None:
            cached_name = self.FN_CACHED%('')
            cleaned_name = self.FN_CLEANED%('')
        else:
            cached_name = self.FN_CACHED%('_' + str(identifier))
            cleaned_name = self.FN_CLEANED%('_' + str(identifier))

        return self.dataset_folder.joinpath(cached_name),\
               self.dataset_folder.joinpath(cleaned_name)


    def __call__(self, cache=False, keep_original=False, retain_corrections=False, load_cleaned=False, folder_path=None, apply_corrections=True, 
                *args, **kwargs):


        self.retain_corrections = retain_corrections

        # Resolve folders
        self._create_exp_folder()

        ident = kwargs.get('ident')
        fp_cached_dataset, fp_cleaned_dataset = self._gen_filenames_cached(ident) 


        # Only download data if it was not already fetched
        # and no cached version exists that should be loaded
        # and the load_cleaned flag is not set
        if not (self.original_folder.exists() or (fp_cached_dataset.is_file() and cache)
                or folder_path is not None) and not load_cleaned:
            self.original_folder.mkdir(parents=True, exist_ok=True)
            self.downloader.download(self.original_folder)

        elif load_cleaned and not fp_cleaned_dataset.exists():
            fp_cleaned_dataset.parent.mkdir(parents=True, exist_ok=True)
            self.downloader.download_cleaned(fp_cleaned_dataset, ident)


        # Load from cached version if available and cache flag is set
            # or load from cleaned version
            # otherwise load from the function
        if fp_cached_dataset.is_file() and cache and not load_cleaned:
            data = joblib.load(fp_cached_dataset)
        elif fp_cleaned_dataset.is_file() and load_cleaned:
            data = joblib.load(fp_cleaned_dataset)
        else:
            data = self.load_data(folder_path=self.original_folder, **kwargs)

        # Save dataset if downloaded and not already cached
        if cache and not fp_cached_dataset.is_file() and not load_cleaned:
            joblib.dump(data, fp_cached_dataset)

        # Clean up data
        if not cache and fp_cached_dataset.is_file():
            fp_cached_dataset.unlink()
        if not keep_original and self.original_folder.exists():
            _delete_data(self.original_folder)

        # Correct devices or activities
        if apply_corrections:
            self.apply_corrections(data, ident, retain_corrections)

        return data


    @abstractmethod
    def load_data(self, *args, **kwargs):
        raise NotImplementedError









class FeatureImportance(object):
    def __init__(self, md, test_x, test_z):
        self._skater_model, self._skater_interpreter = _create_skater_stuff(md, test_x, test_z)

    def save_plot_feature_importance(self, file_path):
        fig, ax = self._skater_interpreter.feature_importance.plot_feature_importance(
            self._skater_model,
            ascending=True,
            ax=None,
            progressbar=False,
            # model-scoring: difference in log_loss or MAE of training_labels
            # given perturbations. Note this vary rarely makes any significant
            # differences
            method='model-scoring')
        # corss entropy or f1 ('f1', 'cross_entropy')
        #scorer_type='cross_entropy') # type: Figure, axes
        #scorer_type='f1') # type: Figure, axes
        import matplotlib.pyplot as plt
        plt.tight_layout()
        fig.savefig(file_path, dpi=fig.dpi)
        plt.close(fig)


def _create_skater_stuff(mdl, test_x, test_z):
        from skater.model import InMemoryModel
        from skater.core.explanations import Interpretation
        from hassbrain_algorithm.benchmark.interpretation import ModelWrapper
        from hassbrain_algorithm.benchmark.interpretation import _boolean2str

        wrapped_model = ModelWrapper(mdl)
        class_names = mdl.get_state_lbl_lst()
        feature_names = mdl.get_obs_lbl_lst()

        # this has to be done in order for skater to recognize the values as categorical and not numerical
        test_x = _boolean2str(test_x)

        # create interpretation
        interpreter = Interpretation(test_x,
                                     #class_names=class_names,
                                     feature_names=feature_names)

        # create model
        # supports classifiers with or without probability scores
        examples = test_x[:10]
        skater_model = InMemoryModel(wrapped_model.predict,
                                     #target_names=class_names,
                                     feature_names=feature_names,
                                     model_type='classifier',
                                     unique_values=class_names,
                                     probability=False,
                                     examples=examples)

        interpreter.load_data(test_x,
                              training_labels=test_z,
                              feature_names=feature_names)
        # todo flag for deletion (3lines below)
        #    if this can savely be deleted
        tmp = interpreter.data_set.feature_info
        for key, val in tmp.items():
            val['numeric'] = False
        return skater_model, interpreter
